import numpy as np

from kolibri.utils.data_structures import ConfusionMatrix, MOLConfusionMatrix
from kolibri.utils.data_structures import Deque, ComplexDeque


class ClassificationMeasurements():
    """ Class used to keep updated statistics about a classifier, in order
    to be able to provide, at any given moment, any relevant metric about
    that classifier.
    """

    def __init__(self, targets=None, dtype=np.int64):
        super().__init__()
        if targets is not None:
            self.n_targets = len(targets)
        else:
            self.n_targets = 0
        self.confusion_matrix = ConfusionMatrix(self.n_targets, dtype)
        self.last_true_label = None
        self.last_prediction = None
        self.last_sample = None
        self.sample_count = 0
        self.majority_classifier = 0
        self.correct_no_change = 0
        self.targets = targets

    def reset(self):
        if self.targets is not None:
            self.n_targets = len(self.targets)
        else:
            self.n_targets = 0
        self.last_true_label = None
        self.last_prediction = None
        self.last_sample = None
        self.sample_count = 0
        self.majority_classifier = 0
        self.correct_no_change = 0
        self.confusion_matrix.restart(self.n_targets)

    def add_result(self, y_true, y_pred, weight=1.0):
        """ Updates its statistics with the results of a prediction.
        Parameters
        ----------
        y_true: int
            The true label.
        y_pred: int
            The classifier's prediction
        weight: float
            Sample's weight
        """

        true_y = self._get_target_index(y_true, True)
        pred = self._get_target_index(y_pred, True)
        self.confusion_matrix.update(true_y, pred)
        self.sample_count += weight

        if self.get_majority_class() == y_true:
            self.majority_classifier = self.majority_classifier + weight
        if self.last_true_label == y_true:
            self.correct_no_change = self.correct_no_change + weight

        self.last_true_label = y_true
        self.last_prediction = y_pred

    def get_last(self):
        return self.last_true_label, self.last_prediction

    def get_majority_class(self):
        """ Computes the true majority class.
        Returns
        -------
        int
            The true majority class.
        """
        if (self.n_targets is None) or (self.n_targets == 0):
            return False
        majority_class = 0
        max_prob = 0.0
        for i in range(self.n_targets):
            sum_value = 0.0
            for j in range(self.n_targets):
                sum_value += self.confusion_matrix.value_at(i, j)
            sum_value = sum_value / self.sample_count
            if sum_value > max_prob:
                max_prob = sum_value
                majority_class = i

        return majority_class

    def get_accuracy(self):
        """ Computes the accuracy.
        Returns
        -------
        float
            The accuracy.
        """
        sum_value = 0.0
        n, _ = self.confusion_matrix.shape()
        for i in range(n):
            sum_value += self.confusion_matrix.value_at(i, i)
        try:
            accuracy = sum_value / self.sample_count
            print(accuracy)
            return accuracy
        except ZeroDivisionError:
            return 0.0

    def get_incorrectly_classified_ratio(self):
        return 1.0 - self.get_accuracy()

    def _get_target_index(self, target, add_label=False):
        """ Computes the index of an element in the self.targets list.
        Also reshapes the ConfusionMatrix and adds new found targets
        if add is True.
        Parameters
        ----------
        target: int
            A class label.
        add_label: bool
            Either to add new found y_values to the targets list or not.
        Returns
        -------
        int
            The label index in the self.targets list.
        """
        if (self.targets is None) and add_label:
            self.targets = []
            self.targets.append(target)
            self.n_targets = len(self.targets)
            self.confusion_matrix.reshape(len(self.targets), len(self.targets))
        elif (self.targets is None) and (not add_label):
            return None
        if (target not in self.targets) and add_label:
            self.targets.append(target)
            self.n_targets = len(self.targets)
            self.confusion_matrix.reshape(len(self.targets), len(self.targets))
        for i in range(len(self.targets)):
            if self.targets[i] == target:
                return i
        return None

    def get_kappa(self):
        """ Computes the Cohen's kappa coefficient.
        Returns
        -------
        float
            The Cohen's kappa coefficient.
        """
        p0 = self.get_accuracy()
        pc = 0.0
        n_rows, n_cols = self.confusion_matrix.shape()
        for i in range(n_rows):
            row = self.confusion_matrix.row(i)
            column = self.confusion_matrix.column(i)

            sum_row = np.sum(row) / self.sample_count
            sum_column = np.sum(column) / self.sample_count

            pc += sum_row * sum_column
        if pc == 1:
            return 1
        return (p0 - pc) / (1.0 - pc)

    def get_kappa_t(self):
        """ Computes the Cohen's kappa T coefficient. This measures the
        temporal correlation between samples.
        Returns
        -------
        float
            The Cohen's kappa T coefficient.
        """
        p0 = self.get_accuracy()
        if self.sample_count != 0:
            pc = self.correct_no_change / self.sample_count
        else:
            pc = 0
        if pc == 1:
            return 1
        return (p0 - pc) / (1.0 - pc)

    def get_kappa_m(self):
        """ Computes the Cohen's kappa M coefficient.
        Returns
        -------
        float
            The Cohen's kappa M coefficient.
        """
        p0 = self.get_accuracy()
        if self.sample_count != 0:
            pc = self.majority_classifier / self.sample_count
        else:
            pc = 0
        if pc == 1:
            return 1
        return (p0 - pc) / (1.0 - pc)

    @property
    def _matrix(self):
        return self.confusion_matrix.matrix

    def get_info(self):
        return '{}:'.format(type(self).__name__) + \
               ' - sample_count: {}'.format(self.sample_count) + \
               ' - accuracy: {:.6f}'.format(self.get_accuracy()) + \
               ' - kappa: {:.6f}'.format(self.get_kappa()) + \
               ' - kappa_t: {:.6f}'.format(self.get_kappa_t()) + \
               ' - kappa_m: {:.6f}'.format(self.get_kappa_m()) + \
               ' - majority_class: {}'.format(self.get_majority_class())

    def get_class_type(self):
        return 'measurement'


class WindowClassificationMeasurements():
    """ This class will maintain a fixed sized window of the newest information
    about one classifier. It can provide, as requested, any of the relevant
    current metrics about the classifier, measured inside the window.
    To keep track of statistics inside a window, the class will use a
    ConfusionMatrix object, alongside FastBuffers, to simulate fixed sized
    windows of the important classifier's attributes.
    Its functionality is somewhat similar to those of the
    ClassificationMeasurements class. The difference is that the statistics
    kept by this class are local, or partial, while the statistics kept by
    the ClassificationMeasurements class are global.
    At any given moment, it can compute the following statistics: accuracy,
    kappa, kappa_t, kappa_m, majority_class and error rate.
    Parameters
    ----------
    targets: list
        A list containing the possible y_values.
    dtype: texts type (Default: numpy.int64)
        The texts type of the existing y_values.
    window_size: int (Default: 200)
        The width of the window. Determines how many samples the object
        can see.
    Examples
    --------
    """

    def __init__(self, targets=None, dtype=np.int64, window_size=200):
        super().__init__()
        if targets is not None:
            self.n_targets = len(targets)
        else:
            self.n_targets = 0
        self.confusion_matrix = ConfusionMatrix(self.n_targets, dtype)
        self.last_class = None

        self.targets = targets
        self.window_size = window_size
        self.true_labels = Deque(window_size)
        self.predictions = Deque(window_size)
        self.temp = 0
        self.last_prediction = None
        self.last_true_label = None
        self.last_sample = None

        self.majority_classifier = 0
        self.correct_no_change = 0
        self.majority_classifier_correction = Deque(window_size)
        self.correct_no_change_correction = Deque(window_size)

    def reset(self):
        if self.targets is not None:
            self.n_targets = len(self.targets)
        else:
            self.n_targets = 0

        self.true_labels = Deque(self.window_size)
        self.predictions = Deque(self.window_size)
        self.temp = 0
        self.last_prediction = None
        self.last_true_label = None
        self.last_sample = None

        self.majority_classifier = 0
        self.correct_no_change = 0
        self.confusion_matrix.restart(self.n_targets)
        self.majority_classifier_correction = Deque(self.window_size)
        self.correct_no_change_correction = Deque(self.window_size)

    def add_result(self, y_true, y_pred):
        """ Updates its statistics with the results of a prediction.
        If needed it will remove samples from the observation window.
        Parameters
        ----------
        y_true: int
            The true label.
        y_pred: int
            The classifier's prediction
        """
        true_y = self._get_target_index(y_true, True)
        pred = self._get_target_index(y_pred, True)
        old_true = self.true_labels.add_element(np.array([y_true]))
        old_predict = self.predictions.add_element(np.array([y_pred]))

        # Verify if it's needed to decrease the count of any label
        # pair in the confusion matrix
        if (old_true is not None) and (old_predict is not None):
            self.temp += 1
            self.confusion_matrix.remove(self._get_target_index(old_true[0]),
                                         self._get_target_index(old_predict[0]))
            self.correct_no_change += self.correct_no_change_correction.peek()
            self.majority_classifier += self.majority_classifier_correction.peek()

        # Verify if it's needed to decrease the majority_classifier count
        if (self.get_majority_class() == y_true) and (self.get_majority_class() is not None):
            self.majority_classifier += 1
            self.majority_classifier_correction.add_element([-1])
        else:
            self.majority_classifier_correction.add_element([0])

        # Verify if it's needed to decrease the correct_no_change
        if (self.last_true_label == y_true) and (self.last_true_label is not None):
            self.correct_no_change += 1
            self.correct_no_change_correction.add_element([-1])
        else:
            self.correct_no_change_correction.add_element([0])

        self.confusion_matrix.update(true_y, pred)

        self.last_true_label = y_true
        self.last_prediction = y_pred

    def get_last(self):
        return self.last_true_label, self.last_prediction

    def get_majority_class(self):
        """ Computes the window/current true majority class.
        Returns
        -------
        int
            The true window/current majority class.
        """
        if (self.n_targets is None) or (self.n_targets == 0):
            return None
        majority_class = 0
        max_prob = 0.0
        for i in range(self.n_targets):
            sum_value = 0.0
            for j in range(self.n_targets):
                sum_value += self.confusion_matrix.value_at(i, j)
            sum_value = sum_value / self.true_labels.get_current_size()
            if sum_value > max_prob:
                max_prob = sum_value
                majority_class = i

        return majority_class

    def get_accuracy(self):
        """ Computes the window/current accuracy.
        Returns
        -------
        float
            The window/current accuracy.
        """
        sum_value = 0.0
        n, _ = self.confusion_matrix.shape()
        for i in range(n):
            sum_value += self.confusion_matrix.value_at(i, i)
        try:
            accuracy = sum_value / self.true_labels.get_current_size()
            print(accuracy)
            return accuracy
        except ZeroDivisionError:
            return 0.0

    def get_incorrectly_classified_ratio(self):
        return 1.0 - self.get_accuracy()

    def _get_target_index(self, target, add=False):
        """ Computes the index of an element in the self.targets list.
        Also reshapes the ConfusionMatrix and adds new found targets
        if add is True.
        Parameters
        ----------
        target: int
            A class label.
        add: bool
            Either to add new found y_values to the targets list or not.
        Returns
        -------
        int
            The label index in the self.targets list.
        """
        if (self.targets is None) and add:
            self.targets = []
            self.targets.append(target)
            self.n_targets = len(self.targets)
            self.confusion_matrix.reshape(len(self.targets), len(self.targets))
        elif (self.targets is None) and (not add):
            return None
        if target not in self.targets and add:
            self.targets.append(target)
            self.n_targets = len(self.targets)
            self.confusion_matrix.reshape(len(self.targets), len(self.targets))
        for i in range(len(self.targets)):
            if self.targets[i] == target:
                return i
        return None

    def get_kappa(self):
        """ Computes the window/current Cohen's kappa coefficient.
        Returns
        -------
        float
            The window/current Cohen's kappa coefficient.
        """
        p0 = self.get_accuracy()
        pc = 0.0
        n_rows, n_cols = self.confusion_matrix.shape()
        for i in range(n_rows):
            row = self.confusion_matrix.row(i)
            column = self.confusion_matrix.column(i)

            sum_row = np.sum(row) / self.true_labels.get_current_size()
            sum_column = np.sum(column) / self.true_labels.get_current_size()

            pc += sum_row * sum_column

        if pc == 1:
            return 1
        return (p0 - pc) / (1.0 - pc)

    def get_kappa_t(self):
        """ Computes the window/current Cohen's kappa T coefficient. This measures
        the temporal correlation between samples.
        Returns
        -------
        float
            The window/current Cohen's kappa T coefficient.
        """
        p0 = self.get_accuracy()
        if self.sample_count != 0:
            pc = self.correct_no_change / self.sample_count
        else:
            pc = 0
        if pc == 1:
            return 1
        return (p0 - pc) / (1.0 - pc)

    def get_kappa_m(self):
        """ Computes the window/current Cohen's kappa M coefficient.
        Returns
        -------
        float
            The window/current Cohen's kappa M coefficient.
        """
        p0 = self.get_accuracy()
        if self.sample_count != 0:
            pc = self.majority_classifier / self.sample_count
        else:
            pc = 0
        if pc == 1:
            return 1
        return (p0 - pc) / (1.0 - pc)

    @property
    def _matrix(self):
        return self.confusion_matrix.matrix

    @property
    def sample_count(self):
        return self.true_labels.get_current_size()

    def get_class_type(self):
        return 'measurement'

    def get_info(self):
        return '{}:'.format(type(self).__name__) + \
               ' - sample_count: {}'.format(self.sample_count) + \
               ' - window_size: {}'.format(self.window_size) + \
               ' - accuracy: {:.6f}'.format(self.get_accuracy()) + \
               ' - kappa: {:.6f}'.format(self.get_kappa()) + \
               ' - kappa_t: {:.6f}'.format(self.get_kappa_t()) + \
               ' - kappa_m: {:.6f}'.format(self.get_kappa_m()) + \
               ' - majority_class: {}'.format(self.get_majority_class())


class RegressionMeasurements():
    """ This class is used to keep updated statistics over a regression
    learner in a regression problem context.
    It will keep track of global metrics, that can be provided at
    any moment. The relevant metrics kept by an instance of this class
    are: MSE (mean square error) and MAE (mean absolute error).
    """

    def __init__(self):
        super().__init__()
        self.total_square_error = 0.0
        self.average_error = 0.0
        self.sample_count = 0
        self.last_true_label = None
        self.last_prediction = None

    def reset(self):
        self.total_square_error = 0.0
        self.average_error = 0.0
        self.sample_count = 0
        self.last_true_label = None
        self.last_prediction = None

    def add_result(self, y_true, y_pred):
        """ Use the true value and the prediction to update the statistics.
        Parameters
        ----------
        y_true: float
            The true value.
        y_pred: float
            The predicted value.
        """
        self.last_true_label = y_true
        self.last_prediction = y_pred

        self.total_square_error += (y_true - y_pred) * (y_true - y_pred)
        self.average_error += np.absolute(y_true - y_pred)
        self.sample_count += 1

    def get_mean_square_error(self):
        """ Computes the mean square error.
        Returns
        -------
        float
            The mean square error.
        """
        if self.sample_count == 0:
            return 0.0
        else:
            return self.total_square_error / self.sample_count

    def get_average_error(self):
        """ Computes the average error.
        Returns
        -------
        float
            The average error.
        """
        if self.sample_count == 0:
            return 0.0
        else:
            return self.average_error / self.sample_count

    def get_last(self):
        return self.last_true_label, self.last_prediction

    def get_class_type(self):
        return 'measurement'

    def get_info(self):
        return '{}:'.format(type(self).__name__) + \
               ' - sample_count: {}'.format(self.sample_count) + \
               ' - mean_square_error: {:.6f}'.format(self.get_mean_square_error()) + \
               ' - mean_absolute_error: {:.6f}'.format(self.get_average_error())


class WindowRegressionMeasurements():
    """ This class is used to keep updated statistics over a regression
    learner in a regression problem context inside a fixed sized window.
    It uses FastBuffer objects to simulate the fixed sized windows.
    It will keep track of partial metrics, that can be provided at
    any moment. The relevant metrics kept by an instance of this class
    are: MSE (mean square error) and MAE (mean absolute error).
    """

    def __init__(self, window_size=200):
        super().__init__()
        self.total_square_error = 0.0
        self.average_error = 0.0
        self.last_true_label = None
        self.last_prediction = None
        self.total_square_error_correction = Deque(window_size)
        self.average_error_correction = Deque(window_size)
        self.window_size = window_size

    def reset(self):
        self.total_square_error = 0.0
        self.average_error = 0.0
        self.last_true_label = None
        self.last_prediction = None
        self.total_square_error_correction = Deque(self.window_size)
        self.average_error_correction = Deque(self.window_size)

    def add_result(self, y_true, y_pred):
        """ Use the true value and the prediction to update the statistics.
        Parameters
        ----------
        y_true: float
            The true value.
        y_pred: float
            The predicted value.
        """
        self.last_true_label = y_true
        self.last_prediction = y_pred
        self.total_square_error += (y_true - y_pred) * (y_true - y_pred)
        self.average_error += np.absolute(y_true - y_pred)

        old_square = self.total_square_error_correction.add_element(
            np.array([-1 * ((y_true - y_pred) * (y_true - y_pred))]))
        old_average = self.average_error_correction.add_element(np.array([-1 * (np.absolute(y_true - y_pred))]))

        if (old_square is not None) and (old_average is not None):
            self.total_square_error += old_square[0]
            self.average_error += old_average[0]

    def get_mean_square_error(self):
        """ Computes the window/current mean square error.
        Returns
        -------
        float
            The window/current mean square error.
        """
        if self.sample_count == 0:
            return 0.0
        else:
            return self.total_square_error / self.sample_count

    def get_average_error(self):
        """ Computes the window/current average error.
        Returns
        -------
        float
            The window/current average error.
        """
        if self.sample_count == 0:
            return 0.0
        else:
            return self.average_error / self.sample_count

    def get_last(self):
        return self.last_true_label, self.last_prediction

    @property
    def sample_count(self):
        return self.total_square_error_correction.get_current_size()

    def get_class_type(self):
        return 'measurement'

    def get_info(self):
        return '{}:'.format(type(self).__name__) + \
               ' - sample_count: {}'.format(self.sample_count) + \
               ' - mean_square_error: {:.6f}'.format(self.get_mean_square_error()) + \
               ' - mean_absolute_error: {:.6f}'.format(self.get_average_error())


class MultiTargetClassificationMeasurements():
    """ This class will keep updated statistics about a multi output classifier,
    using a confusion matrix adapted to multi output problems, the
    MOLConfusionMatrix, alongside other relevant attributes.

    The performance metrics for multi output tasks are different from those used
    for normal classification tasks. Thus, the statistics provided by this class
    are different from those provided by the ClassificationMeasurements and from
    the WindowClassificationMeasurements.

    At any given moment, it can compute the following statistics: hamming_loss,
    hamming_score, exact_match and j_index.

    Parameters
    ----------
    targets: list
        A list containing the possible y_values.

    dtype: texts type (Default: numpy.int64)
        The texts type of the existing y_values.

    Examples
    --------

    """

    def __init__(self, targets=None, dtype=np.int64):
        super().__init__()
        if targets is not None:
            self.n_targets = len(targets)
        else:
            self.n_targets = 0
        self.confusion_matrix = MOLConfusionMatrix(self.n_targets, dtype)
        self.last_true_label = None
        self.last_prediction = None
        self.sample_count = 0
        self.targets = targets
        self.exact_match_count = 0
        self.j_sum = 0

    def reset(self):
        if self.targets is not None:
            self.n_targets = len(self.targets)
        else:
            self.n_targets = 0
        self.confusion_matrix.restart(self.n_targets)
        self.last_true_label = None
        self.last_prediction = None
        self.sample_count = 0
        self.exact_match_count = 0
        self.j_sum = 0

    def add_result(self, y_true, y_pred):
        """ Updates its statistics with the results of a prediction.

        Adds the result to the MOLConfusionMatrix and update exact_matches and
        j-index sum counts.

        Parameters
        ----------
        y_true: list or numpy.ndarray
            The true label.

        y_pred: list or numpy.ndarray
            The classifier's prediction

        """
        self.last_true_label = y_true
        self.last_prediction = y_pred
        m = 0
        if isinstance(y_true, np.ndarray):
            m = y_true.size
        elif isinstance(y_true, list):
            m = len(y_true)
        self.n_targets = m
        equal = True
        for i in range(m):
            self.confusion_matrix.update(i, y_true[i], y_pred[i])
            # update exact_match count
            if y_true[i] != y_pred[i]:
                equal = False

        # update exact_match
        if equal:
            self.exact_match_count += 1

        # update j_index count
        inter = sum((y_true * y_pred) > 0) * 1.
        union = sum((y_true + y_pred) > 0) * 1.
        if union > 0:
            self.j_sum += inter / union
        elif np.sum(y_true) == 0:
            self.j_sum += 1

        self.sample_count += 1

    def get_last(self):
        return self.last_true_label, self.last_prediction

    def get_hamming_loss(self):
        """ Computes the Hamming loss, which is the complement of the
        Hamming score metric.

        Returns
        -------
        float
            The hamming loss.

        """
        return 1.0 - self.get_hamming_score()

    def get_hamming_score(self):
        """ Computes the Hamming score, defined as the number of correctly
        classified y_values divided by the total number of y_values classified.

        Returns
        -------
        float
            The Hamming score.

        """
        try:
            return self.confusion_matrix.get_sum_main_diagonal() / (self.sample_count * self.n_targets)
        except ZeroDivisionError:
            return 0.0

    def get_exact_match(self):
        """ Computes the exact match metric.

        This is the most strict multi output metric, defined as the number of
        samples that have all their y_values correctly classified, divided by the
        total number of samples.

        Returns
        -------
        float
            The exact match metric.

        """
        return self.exact_match_count / self.sample_count

    def get_j_index(self):
        """ Computes the Jaccard index, also known as the intersection over union
        metric. It is calculated by dividing the number of correctly classified
        y_values by the union of predicted and true y_values.

        Returns
        -------
        float
            The Jaccard index.

        """
        return self.j_sum / self.sample_count

    def get_total_sum(self):
        return self.confusion_matrix.get_total_sum()

    @property
    def _matrix(self):
        return self.confusion_matrix.matrix

    def get_info(self):
        return '{}:'.format(type(self).__name__) + \
               ' - sample_count: {}'.format(self.sample_count) + \
               ' - hamming_loss: {:.6f}'.format(self.get_hamming_loss()) + \
               ' - hamming_score: {:.6f}'.format(self.get_hamming_score()) + \
               ' - exact_match: {:.6f}'.format(self.get_exact_match()) + \
               ' - j_index: {:.6f}'.format(self.get_j_index())

    def get_class_type(self):
        return 'measurement'


class WindowMultiTargetClassificationMeasurements():
    """ This class will maintain a fixed sized window of the newest information
    about one classifier. It can provide, as requested, any of the relevant
    current metrics about the classifier, measured inside the window.

    This class will keep updated statistics about a multi output classifier,
    using a confusion matrix adapted to multi output problems, the
    MOLConfusionMatrix, alongside other of the classifier's relevant
    attributes stored in ComplexFastBuffer objects, which will simulate
    fixed sized windows.

    Its functionality is somewhat similar to those of the
    MultiTargetClassificationMeasurements class. The difference is that the statistics
    kept by this class are local, or partial, while the statistics kept by
    the MultiTargetClassificationMeasurements class are global.

    At any given moment, it can compute the following statistics: hamming_loss,
    hamming_score, exact_match and j_index.

    """

    def __init__(self, targets=None, dtype=np.int64, window_size=200):
        super().__init__()
        if targets is not None:
            self.n_targets = len(targets)
        else:
            self.n_targets = 0
        self.confusion_matrix = MOLConfusionMatrix(self.n_targets, dtype)
        self.last_true_label = None
        self.last_prediction = None

        self.targets = targets
        self.window_size = window_size
        self.exact_match_count = 0
        self.j_sum = 0
        self.true_labels = ComplexDeque(window_size, self.n_targets)
        self.predictions = ComplexDeque(window_size, self.n_targets)

    def reset(self):
        if self.targets is not None:
            self.n_targets = len(self.targets)
        else:
            self.n_targets = 0
        self.confusion_matrix.restart(self.n_targets)
        self.last_true_label = None
        self.last_prediction = None
        self.exact_match_count = 0
        self.j_sum = 0
        self.true_labels = ComplexDeque(self.window_size, self.n_targets)
        self.predictions = ComplexDeque(self.window_size, self.n_targets)

    def add_result(self, y_true, y_pred):
        """ Updates its statistics with the results of a prediction.

        Adds the result to the MOLConfusionMatrix, and updates the
        ComplexFastBuffer objects.

        Parameters
        ----------
        y_true: list or numpy.ndarray
            The true label.

        y_pred: list or numpy.ndarray
            The classifier's prediction

        """
        self.last_true_label = y_true
        self.last_prediction = y_pred
        m = 0
        if hasattr(y_true, 'size'):
            m = y_true.size
        elif hasattr(y_true, 'append'):
            m = len(y_true)
        self.n_targets = m

        for i in range(m):
            self.confusion_matrix.update(i, y_true[i], y_pred[i])

        old_true = self.true_labels.add_element(y_true)
        old_predict = self.predictions.add_element(y_pred)
        if (old_true is not None) and (old_predict is not None):
            for i in range(m):
                self.confusion_matrix.remove(old_true[0][i], old_predict[0][i])

    def get_last(self):
        return self.last_true_label, self.last_prediction

    def get_hamming_loss(self):
        """ Computes the window/current Hamming loss, which is the
        complement of the Hamming score metric.

        Returns
        -------
        float
            The window/current hamming loss.

        """
        return 1.0 - self.get_hamming_score()

    def get_hamming_score(self):
        """ Computes the window/current Hamming score, defined as the number of
        correctly classified y_values divided by the total number of y_values
        classified.

        Returns
        -------
        float
            The window/current hamming score.

        """
        return hamming_score(self.true_labels.get_queue(), self.predictions.get_queue())

    def get_exact_match(self):
        """ Computes the window/current exact match metric.

        This is the most strict multi output metric, defined as the number of
        samples that have all their y_values correctly classified, divided by the
        total number of samples.

        Returns
        -------
        float
            The window/current exact match metric.

        """
        return exact_match(self.true_labels.get_queue(), self.predictions.get_queue())

    def get_j_index(self):
        """ Computes the window/current Jaccard index, also known as the intersection
        over union metric. It is calculated by dividing the number of correctly
        classified y_values by the union of predicted and true y_values.

        Returns
        -------
        float
            The window/current Jaccard index.

        """
        return j_index(self.true_labels.get_queue(), self.predictions.get_queue())

    def get_total_sum(self):
        return self.confusion_matrix.get_total_sum()

    @property
    def matrix(self):
        return self.confusion_matrix.matrix

    @property
    def sample_count(self):
        return self.true_labels.get_current_size()

    def get_info(self):
        return '{}:'.format(type(self).__name__) + \
               ' - sample_count: {}'.format(self.sample_count) + \
               ' - hamming_loss: {:.6f}'.format(self.get_hamming_loss()) + \
               ' - hamming_score: {:.6f}'.format(self.get_hamming_score()) + \
               ' - exact_match: {:.6f}'.format(self.get_exact_match()) + \
               ' - j_index: {:.6f}'.format(self.get_j_index())

    def get_class_type(self):
        return 'measurement'


class MultiTargetRegressionMeasurements():
    """ This class is used to keep updated statistics over a multi-label regression
    learner in a multi-label regression problem context.

    It will keep track of global metrics, that can be provided at
    any moment. The relevant metrics kept by an instance of this class
    are: AMSE (average mean square error) and AMAE (average mean absolute error).

    """

    def __init__(self):
        super().__init__()
        self.n_targets = 0
        self.total_square_error = 0.0
        self.average_error = 0.0
        self.sample_count = 0
        self.last_true_label = None
        self.last_prediction = None

    def reset(self):
        self.total_square_error = 0.0
        self.average_error = 0.0
        self.sample_count = 0
        self.last_true_label = None
        self.last_prediction = None

    def add_result(self, y, prediction):
        """ Use the true value and the prediction to update the statistics.

        Parameters
        ----------
        y: float or list or np.ndarray
            The true value(s).

        prediction: float or list or np.ndarray
            The predicted value(s).

        """
        self.last_true_label = y
        self.last_prediction = prediction

        m = 0
        if hasattr(y, 'size'):
            m = y.size
        elif hasattr(y, 'append'):
            m = len(y)
        self.n_targets = m

        self.total_square_error += (y - prediction) ** 2
        self.average_error += np.absolute(y - prediction)
        self.sample_count += 1

    def get_average_mean_square_error(self):
        """ Computes the average mean square error.

        Returns
        -------
        float
            The average mean square error.

        """
        if self.sample_count == 0:
            return 0.0
        else:
            return np.sum(self.total_square_error / self.sample_count) / self.n_targets

    def get_average_absolute_error(self):
        """ Computes the average mean absolute error.

        Returns
        -------
        float
            The average absolute error.

        """
        if self.sample_count == 0:
            return 0.0
        else:
            return np.sum(self.average_error / self.sample_count) \
                   / self.n_targets

    def get_average_root_mean_square_error(self):
        """ Computes the mean square error.

        Returns
        -------
        float
            The average mean square error.

        """
        if self.sample_count == 0:
            return 0.0
        else:
            return np.sum(np.sqrt(self.total_square_error /
                                  self.sample_count)) \
                   / self.n_targets

    def get_last(self):
        return self.last_true_label, self.last_prediction

    @property
    def _sample_count(self):
        return self.sample_count

    def get_class_type(self):
        return 'measurement'

    def get_info(self):
        return 'MultiTargetRegressionMeasurements: sample_count: ' + \
               str(self._sample_count) + ' - average_mean_square_error: ' + \
               str(self.get_average_mean_square_error()) + ' - average_mean_absolute_error: ' + \
               str(self.get_average_absolute_error()) + ' - average_root_mean_square_error: ' + \
               str(self.get_average_root_mean_square_error())


class WindowMultiTargetRegressionMeasurements():
    """ This class is used to keep updated statistics over a multi-label regression
    learner in a multi-label regression problem context inside a fixed sized
    window. It uses FastBuffer objects to simulate the fixed sized windows.

    It will keep track of partial metrics, that can be provided at
    any moment. The relevant metrics kept by an instance of this class
    are: AMSE (average mean square error) and AMAE (average mean absolute error).

    """

    def __init__(self, window_size=200):
        super().__init__()
        self.n_targets = 0
        self.total_square_error = 0.0
        self.average_error = 0.0
        self.last_true_label = None
        self.last_prediction = None
        self.total_square_error_correction = Deque(window_size)
        self.average_error_correction = Deque(window_size)
        self.window_size = window_size

    def reset(self):
        self.total_square_error = 0.0
        self.average_error = 0.0
        self.last_true_label = None
        self.last_prediction = None
        self.total_square_error_correction = Deque(self.window_size)
        self.average_error_correction = Deque(self.window_size)

    def add_result(self, y, prediction):
        """ Use the true value and the prediction to update the statistics.

        Parameters
        ----------
        y: float or list or np.ndarray
            The true value(s).

        prediction: float or list or np.ndarray
            The predicted value(s).

        """
        self.last_true_label = y
        self.last_prediction = prediction

        m = 0
        if hasattr(y, 'size'):
            m = y.size
        elif hasattr(y, 'append'):
            m = len(y)
        self.n_targets = m

        self.total_square_error += (y - prediction) ** 2
        self.average_error += np.absolute(y - prediction)

        old_square = self.total_square_error_correction.add_element(
            np.array([-1 * ((y - prediction) ** 2)])
        )
        old_average = self.average_error_correction.add_element(
            np.array([-1 * (np.absolute(y - prediction))])
        )

        if (old_square is not None) and (old_average is not None):
            self.total_square_error += old_square[0]
            self.average_error += old_average[0]

    def get_average_mean_square_error(self):
        """ Computes the window/current average mean square error.

        Returns
        -------
        float
            The window/current average mean square error.

        """
        if self._sample_count == 0:
            return 0.0
        else:
            return np.sum(self.total_square_error / self._sample_count) \
                   / self.n_targets

    def get_average_absolute_error(self):
        """ Computes the window/current average mean absolute error.

        Returns
        -------
        float
            The window/current average mean absolute error.

        """
        if self._sample_count == 0:
            return 0.0
        else:
            return np.sum(self.average_error / self._sample_count) \
                   / self.n_targets

    def get_average_root_mean_square_error(self):
        """ Computes the mean square error.

        Returns
        -------
        float
            The average mean square error.

        """
        if self._sample_count == 0:
            return 0.0
        else:
            return np.sum(np.sqrt(self.total_square_error /
                                  self._sample_count)) \
                   / self.n_targets

    def get_last(self):
        return self.last_true_label, self.last_prediction

    @property
    def _sample_count(self):
        return self.total_square_error_correction.get_current_size()

    def get_class_type(self):
        return 'measurement'

    def get_info(self):
        return 'MultiTargetRegressionMeasurements: sample_count: ' + \
               str(self._sample_count) + ' - average_mean_square_error: ' + \
               str(self.get_average_mean_square_error()) + ' - average_mean_absolute_error: ' + \
               str(self.get_average_absolute_error()) + ' - average_root_mean_square_error: ' + \
               str(self.get_average_root_mean_square_error())


def hamming_score(true_labels, predicts):
    """ Computes de hamming score, which is known as the label-based accuracy,
    designed for multi-label problems. It's defined as the number of correctly
    predicted y_values divided by all classified y_values.
    """
    if not hasattr(true_labels, 'shape'):
        true_labels = np.asarray(true_labels)
    if not hasattr(predicts, 'shape'):
        predicts = np.asarray(predicts)
    N, L = true_labels.shape
    return np.sum((true_labels == predicts) * 1.) / N / L


def j_index(true_labels, predicts):
    """ Computes the Jaccard Index of the given set, which is also called the
    'intersection over union' in multi-label settings. It's defined as the
    intersection between the true label's set and the prediction's set,
    divided by the sum, or union, of those two sets.
    Parameters
    ----------
    true_labels: numpy.ndarray of shape (n_samples, n_target_tasks)
        A matrix with the true y_values for all the classification tasks and for
        n_samples.
    predicts: numpy.ndarray of shape (n_samples, n_target_tasks)
        A matrix with the predictions for all the classification tasks and for
        n_samples.
    Returns
    -------
    float
        The J-index, or 'intersection over union', for the given sets.
    """
    if not hasattr(true_labels, 'shape'):
        true_labels = np.asarray(true_labels)
    if not hasattr(predicts, 'shape'):
        predicts = np.asarray(predicts)
    N, L = true_labels.shape
    s = 0.0
    for i in range(N):
        inter = sum((true_labels[i, :] * predicts[i, :]) > 0) * 1.
        union = sum((true_labels[i, :] + predicts[i, :]) > 0) * 1.
        if union > 0:
            s += inter / union
        elif np.sum(true_labels[i, :]) == 0:
            s += 1.
    return s * 1. / N


def exact_match(true_labels, predicts):
    """ This is the most strict metric for the multi label setting. It's defined
    as the percentage of samples that have all their y_values correctly classified.
    Parameters
    ----------
    true_labels: numpy.ndarray of shape (n_samples, n_target_tasks)
        A matrix with the true y_values for all the classification tasks and for
        n_samples.
    predicts: numpy.ndarray of shape (n_samples, n_target_tasks)
        A matrix with the predictions for all the classification tasks and for
        n_samples.
    Returns
    -------
    float
        The exact match percentage between the given sets.
    """
    if not hasattr(true_labels, 'shape'):
        true_labels = np.asarray(true_labels)
    if not hasattr(predicts, 'shape'):
        predicts = np.asarray(predicts)
    N, L = true_labels.shape
    return np.sum(np.sum((true_labels == predicts) * 1, axis=1) == L) * 1. / N



import datetime
import os

from kolibri import utils
from kolibri import version as ver
from kolibri.config import component_config_from_pipeline
from kolibri.errors import *


class Metadata(object):
    """Captures all information about a model to load and prepare it."""

    @staticmethod
    def load(model_dir):
        """Loads the metadata from a models directory.

        Args:
            model_dir (str): the directory where the model is saved.
        Returns:
            Metadata: A metadata object describing the model
        """
        try:
            metadata_file = os.path.join(model_dir, 'metadata.json')
            data = utils.file.read_json_file(metadata_file)
            return Metadata(data, model_dir)
        except Exception as e:
            abspath = os.path.abspath(os.path.join(model_dir, 'metadata.json'))
            raise InvalidProjectError("Failed to load model metadata "
                                      "from '{}'. {}".format(abspath, e))

    def __init__(self, metadata, model_dir):

        self.metadata = metadata
        self.model_dir = model_dir

    def get(self, property_name, default=None):
        return self.metadata.get(property_name, default)

    @property
    def component_classes(self):
        if self.get('pipeline'):
            return [c.get("label") for c in self.get('pipeline', [])]
        else:
            return []

    def for_component(self, name, defaults=None):
        return component_config_from_pipeline(name,
                                              self.get('pipeline', []),
                                              defaults)

    @property
    def language(self):
        """Language of the underlying model"""

        return self.get('language')

    def persist(self, model_dir):
        """Persists the metadata of a model to a given directory."""

        metadata = self.metadata.copy()

        metadata.update({
            "trained_at": datetime.datetime.now().strftime('%Y%m%d-%H%M%S'),
            "kolibri_nlu_version": ver.__version__,
        })

        filename = os.path.join(model_dir, 'metadata.json')
        utils.write_json_to_file(filename, metadata, indent=4)



class PostProcessorBase(object):
    """
    A superclass for all adapter classes.

    :param application: An App instance.
    """

    def __init__(self, application, **kwargs):
        self.application = application



from kolibri.tokenizer.char_tokenizer import CharTokenizer
from kolibri.tokenizer.regex_tokenizer import RegexpTokenizer
from kolibri.tokenizer.sentence_tokenizer import SentenceTokenizer
from kolibri.tokenizer.word_tokenizer import WordTokenizer



from __future__ import print_function, unicode_literals

from functools import reduce

import numpy as np
from gensim.models import Word2Vec
from sklearn.preprocessing import StandardScaler

from kolibri.settings import EMBEDDING_SIZE, WORD2VEC_WORKERS, MIN_WORD_COUNT, WORD2VEC_CONTEXT
from kolibri.utils import chunks
from kolibri.utils import flatten
from kolibri.utils.file import save_to_disk


def train_word2vec_in_memory(docs, vec_dim=EMBEDDING_SIZE):
    """
    Builds word embeddings from documents and return a model
    :param docs: list of Document objects
    :param vec_dim: the dimensionality of the vector that's being built

    :return: trained gensim object with word embeddings
    """
    doc_sentences = map(lambda d: d.read_sentences(), docs)
    all_sentences = reduce(lambda d1, d2: d1 + d2, doc_sentences)

    # Initialize and train the model
    model = Word2Vec(
        all_sentences,
        workers=WORD2VEC_WORKERS,
        size=vec_dim,
        min_count=MIN_WORD_COUNT,
        window=WORD2VEC_CONTEXT,
    )

    # If you don't plan to train the model any further, calling
    # init_sims will make the model much more memory-efficient.
    model.init_sims(replace=True)

    return model


def compute_word2vec_for_phrase(phrase, model):
    """
    Compute (add) word embedding for a multiword phrase using a given model
    :param phrase: unicode, parsed label of a keyphrase
    :param model: gensim word2vec object

    :return: numpy array
    """
    result = np.zeros(model.vector_size, dtype='float32')
    for word in phrase.split():
        if word in model.wv:
            result += model.wv[word]

    return result


def fit_scaler(datastream, word2vec_model, batch_size=1024, persist_to_path=None):
    """ Get all the word2vec vectors in a 2D matrix and fit the scaler on it.
     This scaler can be used afterwards for normalizing feature matrices. """
    if type(word2vec_model) == str:
        word2vec_model = Word2Vec.load(word2vec_model)

    scaler = StandardScaler(copy=False)

    for batch in chunks([d.sentences for d in datastream], batch_size):
        vectors = []
        for doc in flatten(batch):
            for word in doc:
                if word in word2vec_model.wv:
                    vectors.append(word2vec_model.wv[word])

        matrix = np.array(vectors)
        print("Fitted to {} vectors".format(matrix.shape[0]))

        scaler.partial_fit(matrix)

    if persist_to_path:
        save_to_disk(persist_to_path, scaler)

    return scaler


def train_word2vec(training_data, vec_dim=EMBEDDING_SIZE):
    """
    Train the Word2Vec object iteratively, loading stuff to memory one by one.
    :param doc_directory: directory with the documents
    :param vec_dim: the dimensionality of the vector that's being built

    :return: Word2Vec object
    """

    class SentenceIterator(object):
        def __init__(self, data):
            self.training_data = data

        def __iter__(self):

            for document in self.training_data:
                for sentence in document.sentences:
                    yield sentence

    # Initialize and train the model
    model = Word2Vec(
        SentenceIterator(training_data),
        workers=WORD2VEC_WORKERS,
        size=vec_dim,
        min_count=MIN_WORD_COUNT,
        window=WORD2VEC_CONTEXT,
    )

    # If you don't plan to train the model any further, calling
    # init_sims will make the model much more memory-efficient.
    model.init_sims(replace=True)

    return model



from seqeval import metrics as seq_metrics
from sklearn import metrics
from tensorflow import keras

from kolibri import config
from kolibri.dnn.tasks.text.base_model import TextBaseModel


class EvalCallBack(keras.callbacks.Callback):

    def __init__(self, kolibri_model: TextBaseModel, valid_x, valid_y,
                 step=5, batch_size=256, average='weighted'):
        """
        Evaluate callback, calculate precision, recall and f1
        Args:
            kolibri_model: the kolibri.dnn model to evaluate
            valid_x: feature texts
            valid_y: label texts
            step: step, default 5
            batch_size: batch size, default 256
        """
        super(EvalCallBack, self).__init__()
        self.kolibri_model = kolibri_model
        self.valid_x = valid_x
        self.valid_y = valid_y
        self.step = step
        self.batch_size = batch_size
        self.average = average
        self.logs = []

    def on_epoch_end(self, epoch, logs=None):
        if (epoch + 1) % self.step == 0:
            y_pred = self.kolibri_model.predict(self.valid_x, batch_size=self.batch_size)

            if self.kolibri_model.task == config.TaskType.LABELING:
                y_true = [seq[:len(y_pred[index])] for index, seq in enumerate(self.valid_y)]
                precision = seq_metrics.precision_score(y_true, y_pred)
                recall = seq_metrics.recall_score(y_true, y_pred)
                f1 = seq_metrics.f1_score(y_true, y_pred)
            else:
                y_true = self.valid_y
                precision = metrics.precision_score(y_true, y_pred, average=self.average)
                recall = metrics.recall_score(y_true, y_pred, average=self.average)
                f1 = metrics.f1_score(y_true, y_pred, average=self.average)

            self.logs.append({
                'precision': precision,
                'recall': recall,
                'f1': f1
            })
            print(f"\nepoch: {epoch} precision: {precision:.6f}, recall: {recall:.6f}, f1: {f1:.6f}")


if __name__ == "__main__":
    print("Hello world")



DATA_POINTS = 'data_points'
ACCURACY = 'accuracy'
F1 = 'f1'
HAMMING_SCORE = 'hamming_score'
HAMMING_LOSS = 'hamming_loss'
EXACT_MATCH = 'exact_match'
J_INDEX = 'j_index'
MSE = 'mean_square_error'
MAE = 'mean_absolute_error'
AMSE = 'average_mean_square_error'
AMAE = 'average_mean_absolute_error'
ARMSE = 'average_root_mean_square_error'
TRUE_VS_PREDICTED = 'true_vs_predicted'

PLOT_TYPES = [ACCURACY,
              HAMMING_SCORE,
              HAMMING_LOSS,
              EXACT_MATCH,
              J_INDEX,
              MSE,
              MAE,
              AMSE,
              AMAE,
              ARMSE,
              TRUE_VS_PREDICTED,
              DATA_POINTS]
CLASSIFICATION_METRICS = [ACCURACY,
                          TRUE_VS_PREDICTED,
                          DATA_POINTS]
REGRESSION_METRICS = [MSE,
                      MAE,
                      TRUE_VS_PREDICTED]
MULTI_TARGET_CLASSIFICATION_METRICS = [HAMMING_SCORE,
                                       HAMMING_LOSS,
                                       EXACT_MATCH,
                                       J_INDEX]
MULTI_TARGET_REGRESSION_METRICS = [AMSE,
                                   AMAE,
                                   ARMSE]
CLASSIFICATION = 'classification'
REGRESSION = 'regression'
MULTI_OUTPUT = 'multi_output'
SINGLE_OUTPUT = 'single-output'
MULTI_TARGET_CLASSIFICATION = 'multi_target_classification'
MULTI_TARGET_REGRESSION = 'multi_target_regression'
UNDEFINED = 'undefined'
TASK_TYPES = [CLASSIFICATION,
              REGRESSION,
              SINGLE_OUTPUT,
              MULTI_OUTPUT,
              MULTI_TARGET_CLASSIFICATION,
              MULTI_TARGET_REGRESSION,
              UNDEFINED]
MEAN = 'mean'
CURRENT = 'current'
Y_TRUE = 'y_true'
Y_PRED = 'y_pred'



from imblearn.combine import (
    SMOTEENN, SMOTETomek,
)
from imblearn.over_sampling import (
    SMOTE, ADASYN, BorderlineSMOTE,
)
from imblearn.under_sampling import (
    ClusterCentroids,
    NearMiss,
    RandomUnderSampler,
    EditedNearestNeighbours,
    AllKNN,
    TomekLinks,
    OneSidedSelection,
    CondensedNearestNeighbour,
    NeighbourhoodCleaningRule,
)

from kolibri.data.samplers.auto_smpler import AutoSampler


def get_sampler(sampler, random_state=None):
    '''
    sampler: String
        The method used to perform re-sampling
        currently support: ['RUS', 'CNN', 'ENN', 'NCR', 'Tomek', 'ALLKNN', 'OSS',
            'NM', 'CC', 'SMOTE', 'ADASYN', 'BorderSMOTE', 'SMOTEENN', 'SMOTETomek',
            'ORG', AUTO']
    '''

    if sampler == 'RUS':
        return RandomUnderSampler(random_state=random_state)
    elif sampler == 'CNN':
        return CondensedNearestNeighbour(random_state=random_state)
    elif sampler == 'ENN':
        return EditedNearestNeighbours()
    elif sampler == 'NCR':
        return NeighbourhoodCleaningRule()
    elif sampler == 'Tomek':
        return TomekLinks()
    elif sampler == 'ALLKNN':
        return AllKNN()
    elif sampler == 'OSS':
        return OneSidedSelection(random_state=random_state)
    elif sampler == 'NM':
        return NearMiss()
    elif sampler == 'CC':
        return ClusterCentroids(random_state=random_state)
    elif sampler == 'SMOTE':
        return SMOTE(random_state=random_state)
    elif sampler == 'ADASYN':
        return ADASYN(random_state=random_state)
    elif sampler == 'BorderSMOTE':
        return BorderlineSMOTE(random_state=random_state)
    elif sampler == 'SMOTEENN':
        return SMOTEENN(random_state=random_state)
    elif sampler == 'SMOTETomek':
        return SMOTETomek(random_state=random_state)
    elif sampler == 'AUTO':
        return AutoSampler(random_state=random_state)
    else:
        raise Exception('Unexpected \'by\' type {}'.format(sampler))



# encoding: utf-8

import os
from operator import itemgetter
from pathlib import Path

import pandas as pd
from tensorflow.keras.utils import get_file

from kolibri import settings as k
from kolibri.data.corpus.data_stream import DataStream
from kolibri.data.corpus.file_stream import FileStream

CORPUS_PATH = os.path.join(k.DATA_PATH, 'corpus')

Path(CORPUS_PATH).mkdir(exist_ok=True, parents=True)


class CONLL2003ENCorpus(FileStream):
    __corpus_name__ = 'conll2003_en'
    __zip_file__name = 'https://www.dropbox.com/s/c65bzd23ho73c0l/conll2003.tar.gz?dl=1'

    def __init__(self, subset_name: str = 'train', task_name: str = 'ner'):

        corpus_path = get_file(self.__corpus_name__,
                               self.__zip_file__name,
                               cache_dir=CORPUS_PATH,
                               untar=True)

        if subset_name not in {'train', 'test', 'valid'}:
            raise ValueError()
        self.task_name = task_name
        self.filepath = os.path.join(corpus_path, f'{subset_name}.txt')
        self.data_index = ['pos', 'chunking', 'ner'].index(self.task_name) + 1
        super().__init__(self.filepath, content_col=0, target_cols=self.data_index, filetype='conll')
        if self.task_name not in {'pos', 'chunking', 'ner'}:
            raise ValueError()
        self.prepare()

    def _load_data(self):
        try:
            raw_data = self.read_function(self.filepath)
            self.n_samples = 0
            for line in raw_data:
                self.n_samples += 1
                for token in line:
                    self.target_values.append(itemgetter(*self.target_columns)(token))

            self.target_values = list(self.target_values)

            self.raw_data = self.read_function(self.filepath)

        except FileNotFoundError:
            raise FileNotFoundError("File {} does not exist.".format(self.filepath))
        pass

    def __iter__(self):

        for token in self.raw_data:
            yield [d[self.content_column] for d in token], [d[self.target_columns[0]] for d in token]


class Sentiment140Corpus(FileStream):
    """

    """

    __corpus_name__ = 'Sentiments140'
    __zip_file__name = "https://www.dropbox.com/s/egk8cwupfs05g00/Sentiments140.tar.gz?dl=1"

    def __init__(self, subset_name='sentiment140_sample'):
        self.corpus_path = get_file(self.__corpus_name__,
                                    self.__zip_file__name,
                                    cache_dir=k.DATA_PATH,
                                    untar=True)

        if subset_name not in {'sentiment140_sample', 'all'}:
            raise ValueError()

        self.file_path = os.path.join(self.corpus_path, f'{subset_name}.csv')

        super().__init__(filepath=self.file_path, content_col=' text', target_cols='label')
        self.prepare()

class CreditCardFraud(DataStream):
    """

    """

    __corpus_name__ = 'creditcard_fraud'
    __zip_file__name = "https://www.dropbox.com/s/7v4tm6lsjkxnvfk/creditcard_fraud.tgz?dl=1"

    def __init__(self, subset_name='creditcard'):
        self.corpus_path = get_file(self.__corpus_name__,
                                    self.__zip_file__name,
                                    cache_dir=k.DATA_PATH,
                                    untar=True)

        if subset_name not in {'creditcard'}:
            raise ValueError()

        self.file_path = os.path.join(self.corpus_path, f'{subset_name}.csv')
        data=pd.read_csv(self.file_path)
        columns=[ c for c in data.columns if c not in ['Class', 'Time']]

        super().__init__(data=data[columns], y=data['Class'].values)
        self.prepare()


class ConsumerComplaintsCorpus(FileStream):
    """

    """

    __corpus_name__ = 'consumer_complaints'
    __zip_file__name = "https://www.dropbox.com/s/8a1pm3gg9e5szso/consumer_complaints.tar.gz?dl=1"

    def __init__(self, subset_name='sample'):
        self.corpus_path = get_file(self.__corpus_name__,
                                    self.__zip_file__name,
                                    cache_dir=k.DATA_PATH,
                                    untar=True)

        if subset_name not in {'sample', 'validate', 'train', 'test'}:
            raise ValueError()

        self.file_path = os.path.join(self.corpus_path, f'{subset_name}.csv')

        super().__init__(filepath=self.file_path, content_col='Consumer_complaint', target_cols=['Product'])
        self.prepare()


if __name__ == "__main__":
    corpus = CreditCardFraud()

    for d in corpus.X:
        print(d)




from kolibri.application.postprocssor_base import PostProcessorBase
from kolibri.application.search import IndexedTextSearch
from kolibri.document import Document


class ScorePostProcessor(PostProcessorBase):
    """
    This is an abstract class that represents the interface
    that all score postprocessors should implement.

    :param search_algorithm_name: The name of the search algorithm that should
        be used to search for close matches to the provided input.
        Defaults to the value of ``Search.name``.

    :param maximum_similarity_threshold:
        The maximum amount of similarity between two document that is required
        before the search process is halted. The search for a matching document
        will continue until a document with a greater than or equal similarity
        is found or the search set is exhausted.
        Defaults to 0.95

    :param candidate_selection_method:
          The a response selection method.
          Defaults to ``get_first_response``

    :param default_response:
          The default response returned by this score adaper
          if there is no other possible response to return.
    :type default_response: str or list or tuple
    """

    def __init__(self, application, **kwargs):
        super().__init__(application, **kwargs)
        from kolibri.application.target_selection import get_first_target

        self.search_algorithm_name = kwargs.get(
            'search_algorithm_name',
            IndexedTextSearch.name
        )

        self.search_algorithm = self.application.search_algorithms[
            self.search_algorithm_name
        ]

        self.maximum_similarity_threshold = kwargs.get(
            'maximum_similarity_threshold', 0.95
        )

        # By default, select the first available response
        self.select_response = kwargs.get(
            'candidate_selection_method',
            get_first_target
        )

        default_responses = kwargs.get('default_response', [])

        # Convert a single string into a list
        if isinstance(default_responses, str):
            default_responses = [
                default_responses
            ]

        self.default_responses = [
            Document(text=default) for default in default_responses
        ]

    def can_process(self, document):
        """
        A preliminary check that is called to determine if a
        score postpressor can process a given document. By default,
        this method returns true but it can be overridden in
        child classes as needed.

        :rtype: bool
        """
        return True

    def process(self, document, additional_candidate_selection_parameters=None):
        """
        Override this method and implement your score for selecting a response to an input document.

        A confidence value and the selected response document should be returned.
        The confidence value represents a rating of how accurate the score postprocessor
        expects the selected response to be. Confidence scores are used to select
        the best response from multiple score postprocessors.

        The confidence value should be a number between 0 and 1 where 0 is the
        lowest confidence level and 1 is the highest.

        :param document: An input document to be processed by the score postprocessor.

        :param additional_candidate_selection_parameters: Parameters to be used when
            filtering results to choose a response from.

        :rtype: Document
        """
        raise NotImplementedError

    def get_default_response(self, input_document):
        """
        This method is called when a score postprocessor is unable to generate any
        other meaningful response.
        """
        from random import choice

        if self.default_responses:
            response = choice(self.default_responses)
        else:
            try:
                response = self.application.storage.get_random()
            except Exception:
                response = input_document

        self.application.logger.info(
            'No known response to the input was found. Selecting a random response.'
        )

        # Set confidence to zero because a random response is selected
        response.confidence = 0

        return response

    @property
    def class_name(self):
        """
        Return the name of the current score postprocessor class.
        This is typically used for logging and debugging.
        """
        return str(self.__class__.__name__)



"""
Target selection methods determines which target should be used in
the event that multiple targets are generated within a logic adapter.
"""
import logging

logger = logging.getLogger(__name__)


def get_most_frequent_target(input_document, target_list, storage=None):
    """
    :param input_document: A document, that closely matches an input to the chat bot.
    :type input_document: Statement

    :param target_list: A list of document options to choose a target from.
    :type target_list: list

    :param storage: An instance of a storage adapter to allow the target selection
                    method to access other documents if needed.
    :type storage: StorageAdapter

    :return: The target document with the greatest number of occurrences.
    :rtype: Statement
    """
    matching_target = None
    occurrence_count = -1

    logger.info('Selecting target with greatest number of occurrences.')

    for document in target_list:
        count = len(list(storage.filter(
            text=document.text,
            in_target_to=input_document.text)
        ))

        # Keep the more common document
        if count >= occurrence_count:
            matching_target = document
            occurrence_count = count

    # Choose the most commonly occuring matching target
    return matching_target


def get_first_target(input_document, target_list, storage=None):
    """
    :param input_document: A document, that closely matches an input to the chat bot.
    :type input_document: Statement

    :param target_list: A list of document options to choose a target from.
    :type target_list: list

    :param storage: An instance of a storage adapter to allow the target selection
                    method to access other documents if needed.
    :type storage: StorageAdapter

    :return: Return the first document in the target list.
    :rtype: Statement
    """
    logger = logging.getLogger(__name__)
    logger.info('Selecting first target from list of {} options.'.format(
        len(target_list)
    ))
    return target_list[0]


def get_random_target(input_document, target_list, storage=None):
    """
    :param input_document: A document, that closely matches an input to the chat bot.
    :type input_document: Statement

    :param target_list: A list of document options to choose a target from.
    :type target_list: list

    :param storage: An instance of a storage adapter to allow the target selection
                    method to access other documents if needed.
    :type storage: StorageAdapter

    :return: Choose a random target from the selection.
    :rtype: Statement
    """
    from random import choice
    logger = logging.getLogger(__name__)
    logger.info('Selecting a target from list of {} options.'.format(
        len(target_list)
    ))
    return choice(target_list)



import datetime
import logging
import os
from pathlib import Path

from kolibri import kolibri_component
from kolibri.metadata import Metadata
from kolibri.model_loader import ModelLoader
from kolibri.pipeline import Pipeline
from kolibri.utils import create_dir
from kolibri.utils import module_path_from_object

logger = logging.getLogger(__name__)

MINIMUM_COMPATIBLE_VERSION = "0.0.1"


class ModelTrainer(object):
    """Trainer will load the texts and train all components.

    Requires a pipeline specification and configuration to use for
    the training."""

    SUPPORTED_LANGUAGES = ["de", "en"]

    def __init__(self, cfg, component_builder=None, skip_validation=False):

        self.config = cfg
        self.skip_validation = skip_validation
        self.training_data = None

        if component_builder is None:
            # If no builder is passed, every interpreter creation will result in
            # a new builder. hence, no components are reused.
            component_builder = kolibri_component.ComponentBuilder()

        # Before instantiating the component classes, lets check if all
        # required packages are available
        if not self.skip_validation:
            kolibri_component.validate_requirements(cfg.component_names)

        # build pipeline
        self.pipeline = self._build_pipeline(cfg, component_builder)

    @staticmethod
    def _build_pipeline(cfg, component_builder):
        """Transform the passed names of the pipeline components into classes"""

        steps = []
        # Transform the passed names of the pipeline components into classes
        for component_name in cfg.component_names:
            component = component_builder.create_component(
                component_name, cfg)
            steps.append((component_name, component))

        return Pipeline(steps)

    def fit(self, X, y, X_val=None, y_val=None):
        """Trains the underlying pipeline using the provided training texts."""

        self.pipeline.fit(X, y, X_val, y_val)

        return ModelLoader(self.pipeline, {})

    def train(self, data, **kwargs):
        """Trains the underlying pipeline using the provided training texts."""

        self.training_data = data

        context = kwargs

        for component in self.pipeline.components:
            updates = component[1].provide_context()
            if updates:
                context.update(updates)

        self.pipeline.train(data, **kwargs)

        return ModelLoader(self.pipeline, context)

    def persist(self, path, persistor=None, project_name=None,
                fixed_model_name=None):
        """Persist all components of the pipeline to the passed path.

        Returns the directory of the persisted model."""

        timestamp = datetime.datetime.now().strftime('%Y%m%d-%H%M%S')
        metadata = {
            "language": self.config["language"],
            "pipeline": [],
        }

        if project_name is None:
            project_name = "default"

        if fixed_model_name:
            model_name = fixed_model_name
        else:
            model_name = "model_" + timestamp

        path = Path(path).resolve()
        dir_name = os.path.join(path, project_name, model_name)

        create_dir(dir_name)

        #        if self.training_data:
        #            metadata.update(self.training_data.persist(dir_name))

        for component in self.pipeline.components:
            update = component[1].persist(dir_name)
            component_meta = component[1].component_config
            if update:
                component_meta.update(update)
            component_meta["label"] = module_path_from_object(component[1])
            metadata["pipeline"].append(component_meta)

        Metadata(metadata, dir_name).persist(dir_name)

        if persistor is not None:
            persistor.persist(dir_name, model_name, project_name)
        logger.info("Successfully saved model into "
                    "'{}'".format(os.path.abspath(dir_name)))
        return dir_name



import logging

from sklearn.utils import tosequence

from kolibri.data.corpus.base_stream import BaseStream


class Pipeline():
    """ [Experimental] Holds a set of sequential operation (transforms), followed by a single estimator.
    """
    _estimator_type = 'pipeline'

    def __init__(self, steps=None):
        # Default values
        super().__init__()

        self.steps = tosequence(steps)
        self.active = False

        self.__configure()

    def __configure(self):
        """ __configure
        Initial Pipeline configuration. Validates the Pipeline's steps.
        """
        self._validate_steps()

    def predict(self, X):
        """ predict
        Sequentially applies all transforms and then predict with last step.
        Parameters
        ----------
        X: numpy.ndarray of shape (n_samples, n_features)
            All the samples we want to predict the label for.
        Returns
        -------
        list
            The predicted class label for all the samples in X.
        """
        Xt = X
        for name, transform in self.steps[:-1]:
            if transform is not None:
                Xt = transform.transform(Xt, )
        return self.steps[-1][-1].predict(Xt)

    def predict_prob(self, X):
        """ predict
        Sequentially applies all transforms and then predict with last step.
        Parameters
        ----------
        X: numpy.ndarray of shape (n_samples, n_features)
            All the samples we want to predict the label for.
        Returns
        -------
        list
            The predicted class label for all the samples in X.
        """
        Xt = X
        for name, transform in self.steps[:-1]:
            if transform is not None:
                Xt = transform.transform(Xt, )
        return self.steps[-1][-1].predict_prob(Xt)

    def numerize_label_sequences(self, y):
        return self.steps[-1][-1].indexer.numerize_label_sequences(y)

    @property
    def components(self):
        return self.steps

    def fit(self, X, y, X_val=None, y_val=None):

        """ fit
        Sequentially fit and transformer texts in all but last step, then fit
        the model in last step.
        Parameters
        ----------
        X: numpy.ndarray of shape (n_samples, n_features)
            The texts upon which the transforms/estimator will create their
            model.
        y: An array_like object of length_train n_samples
            Contains the true class y_values for all the samples in X.
        Returns
        -------
        Pipeline
            self
        """
        Xt = X
        Xt_val = X_val
        for name, transformer in self.steps[:-1]:
            if transformer is None:
                pass
            if hasattr(transformer, "fit_transform"):
                Xt = transformer.fit_transform(Xt, y)
                if Xt_val:
                    Xt_val = transformer.fit_transform(Xt_val, y_val)
            else:
                Xt = transformer.fit(Xt, y).transform(Xt, )
                if Xt_val:
                    Xt_val = transformer.transform(Xt_val)

        if self._final_estimator is not None:
            self._final_estimator.fit(Xt, y, Xt_val, y_val)

        return self

    def partial_fit(self, X, y, classes=None):
        """ partial_fit
        Sequentially partial fit and transformer texts in all but last step,
        then partial fit texts in last step.
        Parameters
        ----------
        X : numpy.ndarray of shape (n_samples, n_features)
            The features to train the model.
        y: numpy.ndarray of shape (n_samples)
            An array-like with the class y_values of all samples in X.
        classes: numpy.ndarray
            Array with all possible/known class y_values. This is an optional parameter, except
            for the first partial_fit call where it is compulsory.
        Returns
        -------
        Pipeline
            self
        """
        Xt = X
        for name, transformer in self.steps[:-1]:
            if transformer is None:
                pass
            if hasattr(transformer, 'fit_transform'):
                Xt = transformer.partial_fit_transform(Xt, y, classes=classes)
            elif hasattr(transformer, 'transform'):
                Xt = transformer.transform(Xt)
            else:
                Xt = transformer.partial_fit(Xt, y).transform(Xt, )

        if self._final_estimator is not None:
            if "classes" in self._final_estimator.partial_fit.__code__.co_varnames:
                self._final_estimator.partial_fit(Xt, y, classes)
            else:
                self._final_estimator.partial_fit(Xt, y)
        return self

    def partial_fit_predict(self, X, y):
        """ partial_fit_predict
        Partial fits and transforms texts in all but last step, then partial
        fits and predicts in the last step
        Parameters
        ----------
        X: numpy.ndarray of shape (n_samples, n_features)
            All the samples we want to predict the label for.
        y: An array_like object of length_train n_samples
            Contains the true class y_values for all the samples in X
        Returns
        -------
        list
            The predicted class label for all the samples in X.
        """
        Xt = X
        for name, transform in self.steps[:-1]:
            if transform is None:
                pass
            if hasattr(transform, "partial_fit_transform"):
                Xt = transform.partial_fit_transform(Xt, y)
            else:
                Xt = transform.partial_fit(Xt, y).transform(Xt, )

        if hasattr(self._final_estimator, "partial_fit_predict"):
            return self._final_estimator.partial_fit_predict(Xt, y)
        else:
            return self._final_estimator.partial_fit(Xt, y).predict(Xt)

    def partial_fit_transform(self, X, y=None):
        """ partial_fit_transform
        Partial fits and transforms texts in all but last step, then
        partial_fit in last step
        Parameters
        ----------
        X: numpy.ndarray of shape (n_samples, n_features)
            The texts upon which the transforms/estimator will create their
            model.
        y: An array_like object of length_train n_samples
            Contains the true class y_values for all the samples in X
        Returns
        -------
        Pipeline
            self
        """
        raise NotImplementedError

    def _validate_steps(self):
        """ validate_steps
        Validates all steps, guaranteeing that there's an estimator in its last step.
        Alters the value of self.active according to the validity of the steps.
        Raises
        ------
        TypeError: If the intermediate steps or the final estimator do not implement
        the necessary functions for the pipeline to work, a TypeError is raised.
        """

        names, estimators = zip(*self.steps)
        estimator = estimators[-1]
        transformers = estimators[:-1]

        self.active = True

        for t in transformers:
            if t is None:
                continue
            else:
                if not (hasattr(t, "fit") or hasattr(t, "fit_transform")) or not hasattr(t, "transform"):
                    self.active = False
                    raise TypeError("All intermediate steps, including an evaluator, "
                                    "should implement fit and transform.")

    def stream_fit(self, stream: BaseStream, batch_size=10000):

        logging.basicConfig(format='%(message)s', level=logging.INFO)
        logging.info('Prequential Evaluation')

        while True:
            logging.info('Pre-training on %s samples.', str(batch_size))

            X, y = stream.next(batch_size)
            if len(X) == 0 or len(y) == 0:
                break

            for name, transformer in self.steps[:-1]:
                if hasattr(transformer, "transform"):
                    X = transformer.transform(X, )

            if self._final_estimator is not None:
                self._final_estimator.partial_fit(X, y, stream.target_values)

    def train(self, data, **kwargs):
        """ fit
        Sequentially fit and transform texts in all but last step, then fit
        the model in last step.
        Parameters
        ----------
        training_data: DataStream structure containing training texts

        Returns
        -------
        Pipeline
            self
        """
        training_data = data.get_all_documents()
        for name, transformer in self.steps[:-1]:
            if transformer is None:
                pass
            if hasattr(transformer, "train") and hasattr(transformer, "process"):
                transformer.train(training_data, **kwargs)
            else:
                raise Exception("train or process interface not implemented")

        if self._final_estimator is not None:
            self._final_estimator.train(training_data, **kwargs)

        return self

    def process(self, document, **kwargs):
        """Process an incoming document.
        predict
        Sequentially applies all transforms and then predict with last step.
        Parameters
        ----------
        document: the document to predict.
        Returns
        -------
        list
            The predicted class label for all the samples in X.
        """

        for name, transformer in self.steps[:-1]:
            if transformer is not None:
                transformer.process(document, **kwargs)
        return self.steps[-1][-1].process(document, **kwargs)

    def named_steps(self):
        """ named_steps
        Generates a dictionary to access all the steps' properties.
        Returns
        -------
        dictionary
            A steps dictionary, so that each step can be accessed by name.
        """
        return dict(self.steps)

    def get_info(self):
        info = "Pipeline:\n["
        names, estimators = zip(*self.steps)
        learner = estimators[-1]
        transforms = estimators[:-1]
        i = 0
        for t in transforms:
            try:
                if t.get_info() is not None:
                    info += t.get_info()
                    info += "\n"
                else:
                    info += 'Transform: no info available'
            except NotImplementedError:
                info += 'Transform: no info available'
            i += 1

        if learner is not None:
            try:
                if hasattr(learner, 'get_info'):
                    info += learner.get_info()
                else:
                    info += 'Learner: no info available'
            except NotImplementedError:
                info += 'Learner: no info available'
        info += "]"
        return info

    @property
    def _final_estimator(self):
        """ _final_estimator
        Easy to access estimator.
        Returns
        -------
        Extension of BaseClassifier
            The Pipeline's classifier
        """
        return self.steps[-1][-1]



from kolibri.config import ModelConfig
from kolibri.data.cleaners.audio import AudioPreprocessor
from kolibri.dnn.tasks.audio import *
from kolibri.dnn.tasks.text import *
from kolibri.features import *
from kolibri.kolibri_component import Component
from kolibri.meta.ecoc_estimator import EcocEstimator
from kolibri.meta.ensemble_samplers import *
from kolibri.model_loader import ModelLoader
from kolibri.model_trainer import ModelTrainer
from kolibri.nlp.bigram_analyzer import BigramAnalyzer
from kolibri.tasks.classification import *
from kolibri.tokenizer import *
from kolibri.version import __version__

name = 'kolibri'

#p=Popen(["analyze", "-f /Users/mohamedmentis/Documents/Mentis/Development/Python/Deep_kolibri/kolibri/tools/freeling/config/fr_mwe.cfg","--output", "conll", "--server", "--port", "50101 &"], stdout=subprocess.PIPE,stderr=subprocess.STDOUT)
#time.sleep(10)
#for line in iter(p.stdout.readline,""):
#    print(line)


#"analyze -f /Users/mohamedmentis/Documents/Mentis/Development/Python/Deep_kolibri/kolibri/tools/freeling/config/en_mwe.cfg --output conll --server --port 50101 &"


"""
reformat confusion matrix by prob texts frame;
"""
from typing import Optional

import pandas as pd

_CONFUSION_MATRIX_DIGIT_FORMAT = {'TP': 0,
                                  'FN': 0,
                                  'FP': 0,
                                  'TN': 0,
                                  'Recall': 3,
                                  'FNR': 3,
                                  'FPR': 3,
                                  'TNR': 3,
                                  'Precision': 3,
                                  'FOR': 3,
                                  'FDR': 3,
                                  'NPV': 3,
                                  'Prevalence': 3,
                                  'Accuracy': 3,
                                  'LR+': 3,
                                  'LR-': 3,
                                  'DOR': 3,
                                  'F1': 3}

_DEFAULT_METRIC_ORDER = ['TP', 'FN', 'FP', 'TN',
                         'Recall', 'FNR', 'FPR', 'TNR',
                         'Precision', 'FOR', 'FDR', 'NPV',
                         'Prevalence', 'Accuracy', 'LR+', 'LR-', 'DOR', 'F1',
                         'Open_Count', 'Open_Positive', 'Open_Negative', 'Open_Prevalence']


def round_format(value, round_by):
    if isinstance(value, str):
        pass
    else:
        value = round(value, round_by)
    return value


def _reformat_digit_confusion_matrix_by_prob(metrics_by_thresholds_df: pd.DataFrame,
                                             digit_format: Optional[dict] = None) -> pd.DataFrame:
    """
    reformat the digit representations confusion matrix by prob texts frame by:
        1. finding maximum number of decimal places for threshold series, and set such precision to each threshold,
        2. for format in max_digit_format, apply the setting to each corresponding series;

    :param metrics_by_thresholds_df: pandas.DataFrame, a confusion matrix metric texts frame;
    :param digit_format: dict, format map that gives the precision/decimal places of the series;
    :return: reformatted metric texts frame;
    """

    def get_number_of_decimal_places(number: float) -> int:
        """
        find out how many number of decimal places of a given number;
        :param number: float, a number;
        :return: number of decimal places of the number;
        """
        if number < 0.000001:
            number = 0

        digit_as_str = str(number)

        if '.' not in digit_as_str:
            decimal_no = 0
        else:
            decimal_no = len(digit_as_str.split('.')[1])

        return decimal_no

    if ('l_threshold' in metrics_by_thresholds_df.columns) and ('u_threshold' in metrics_by_thresholds_df.columns):
        threshold_cols = ['l_threshold', 'u_threshold']
    else:
        threshold_cols = ['threshold']

    for threshold_col in threshold_cols:
        max_no_threshold_decimal = max(metrics_by_thresholds_df[threshold_col].apply(get_number_of_decimal_places))

        metrics_by_thresholds_df[threshold_col] = metrics_by_thresholds_df[threshold_col].apply(
            lambda x: round_format(x, max_no_threshold_decimal))

    if not digit_format:
        digit_format = _CONFUSION_MATRIX_DIGIT_FORMAT

    for metric, decimal_format in digit_format.items():
        if metric in metrics_by_thresholds_df.columns:
            metrics_by_thresholds_df[metric] = metrics_by_thresholds_df[metric].apply(
                lambda x: round_format(x, _CONFUSION_MATRIX_DIGIT_FORMAT[metric]))

    return metrics_by_thresholds_df


def _convert_confusion_matrix_by_prob_to_table(metrics_by_thresholds: dict,
                                               metric_order: Optional[list] = None) -> pd.DataFrame:
    """
    reformat the confusion matrix by prob from dict format to texts frame format;

    :param metrics_by_thresholds: a (confusion matrix) metric dict (with other additional metrics) by different thresholds;
    :param metric_order: list, the order by which metrics are presented after threshold series;
        if the item is not in metric order, but presented in metrics_by_thresholds, it will be placed after the last item
            of metric order with random placement;
        if the item is in the metric order but not presented in metrics_by_thresholds, it will be ignored;
    :return: confusion_matrix_by_prob_to_table, pandas.DataFrame;
    """
    _if_dual_threshold = False
    for threshold, metrics in metrics_by_thresholds.items():
        if isinstance(threshold, float):
            metrics['threshold'] = threshold
        elif isinstance(threshold, tuple):
            metrics['l_threshold'], metrics['u_threshold'] = threshold
            if not _if_dual_threshold:  # avoid calculating again and again;
                _if_dual_threshold = True
        else:
            raise ValueError(f"threshold {threshold} is not recognized of it type.")

    if not metric_order:
        metric_order = _DEFAULT_METRIC_ORDER

    if _if_dual_threshold:
        threshold_cols = ['l_threshold', 'u_threshold']
    else:
        threshold_cols = ['threshold']

    metric_order = threshold_cols + [column for column in metric_order if
                                     column in list(metrics_by_thresholds.values())[0]]

    metric_order += [column for column in list(metrics_by_thresholds.values())[0] if column not in metric_order]

    metrics_by_thresholds = list(metrics_by_thresholds.values())
    metrics_by_thresholds_df = pd.DataFrame(metrics_by_thresholds)

    metrics_by_thresholds_df = metrics_by_thresholds_df[metric_order]

    metrics_by_thresholds_df = metrics_by_thresholds_df.sort_values(by=threshold_cols[0], ascending=True)
    metrics_by_thresholds_df = metrics_by_thresholds_df.reset_index(drop=True)

    return metrics_by_thresholds_df


def convert_confusion_matrix_by_prob_to_table_with_reformat_precision(metrics_by_thresholds: dict,
                                                                      metric_order: Optional[list] = None,
                                                                      digit_format: Optional[
                                                                          dict] = None) -> pd.DataFrame:
    """
    wrap function that perform conversion of confusion matrix by prob from dict to pandas.DataFrame with reformatting;

    :param metrics_by_thresholds: dict, read above;
    :param metric_order: list, read above;
    :param digit_format: dict, read above;
    :return: confusion_matrix_by_prob_to_table, pandas.DataFrame;
    """

    metrics_by_thresholds_df = _convert_confusion_matrix_by_prob_to_table(metrics_by_thresholds=metrics_by_thresholds,
                                                                          metric_order=metric_order)

    metrics_by_thresholds_df = _reformat_digit_confusion_matrix_by_prob(
        metrics_by_thresholds_df=metrics_by_thresholds_df,
        digit_format=digit_format)

    return metrics_by_thresholds_df



import logging
from typing import Optional

from kolibri import kolibri_component
from kolibri import version as ver
from kolibri.document import Document
from kolibri.metadata import Metadata
from kolibri.pipeline import Pipeline
from .errors import *
from .kolibri_component import ComponentBuilder

logger = logging.getLogger(__name__)

MINIMUM_COMPATIBLE_VERSION = "0.0.1"


class ModelLoader(object):
    """Use a trained pipeline of components to parse text documents."""

    # Defines all attributes (& default values)
    # that will be returned by `parse`
    @staticmethod
    def default_output_attributes():
        return {"label": {"name": None, "confidence": 0.0}, "entities": []}

    @staticmethod
    def ensure_model_compatibility(metadata, version_to_check=None):
        from packaging import version

        if version_to_check is None:
            version_to_check = MINIMUM_COMPATIBLE_VERSION

        model_version = metadata.get("kolibri_nlu_version", "0.0.0")
        if version.parse(model_version) < version.parse(version_to_check):
            raise UnsupportedModelError(
                "The model version is to old to be "
                "loaded by this nlp instance. "
                "Either retrain the model, or run with"
                "an older version. "
                "Model version: {} Instance version: {}"
                "".format(model_version, ver.__version__))

    @staticmethod
    def load(model_dir, component_builder=None, skip_validation=False):
        """Create an interpreter based on a persisted model.

        Args:
            model_dir (str): The path of the model to load
            component_builder (ComponentBuilder): The
                :class:`ComponentBuilder` to use.

        Returns:
            ModelLoader: An interpreter that uses the loaded model.
        """

        model_metadata = Metadata.load(model_dir)

        ModelLoader.ensure_model_compatibility(model_metadata)
        return ModelLoader.create(model_metadata,
                                  component_builder,
                                  skip_validation)

    @staticmethod
    def create(model_metadata,  # type: Metadata
               component_builder=None,  # type: Optional[ComponentBuilder]
               skip_validation=False  # type: bool
               ):
        # type: (...) -> ModelLoader
        """Load stored model and components defined by the provided metadata."""

        context = {}

        if component_builder is None:
            # If no builder is passed, every interpreter creation will result
            # in a new builder. hence, no components are reused.
            component_builder = kolibri_component.ComponentBuilder()

        steps = []

        # Before instantiating the component classes,
        # lets check if all required packages are available
        if not skip_validation:
            kolibri_component.validate_requirements(model_metadata.component_classes)

        for component_name in model_metadata.component_classes:
            component = component_builder.load_component(
                component_name, model_metadata.model_dir,
                model_metadata, **context)
            try:
                updates = component.provide_context()
                if updates:
                    context.update(updates)
                steps.append((component.name, component))
            except kolibri_component.MissingArgumentError as e:
                raise Exception("Failed to initialize component '{}'. "
                                "{}".format(component.name, e))

        return ModelLoader(Pipeline(steps), context, model_metadata)

    def __init__(self, pipeline: Pipeline, context, model_metadata=None):
        """

        :type pipeline: Pipeline
        """
        self.pipeline = pipeline
        self.context = context if context is not None else {}
        self.model_metadata = model_metadata

    def predict(self, text):
        """Predict the input text, classify it and return pipeline result.

        The pipeline result usually contains intent and entities."""

        if text is None:
            # Not all components are able to handle empty strings. So we need
            # to prevent that... This default return will not contain all
            # output attributes of all components, but in the end, no one
            # should pass an empty string in the first place.
            output = self.default_output_attributes()
            output["text"] = ""
            return output

        return self.pipeline.predict(text)

    def process(self, text, time=None, only_output_properties=True):
        """Predict the input text, classify it and return pipeline result.

        The pipeline result usually contains intent and entities."""

        if not text:
            # Not all components are able to handle empty strings. So we need
            # to prevent that... This default return will not contain all
            # output attributes of all components, but in the end, no one
            # should pass an empty string in the first place.
            output = self.default_output_attributes()
            output["text"] = ""
            return output

        document = Document(**{'text':text, 'default_output_attributes': self.default_output_attributes(), 'time':time})

        self.pipeline.process(document, **self.context)

        output = self.default_output_attributes()
        output.update(document.as_dict(
            only_output_properties=only_output_properties))

        return output

    def get_component(self, component):
        for c in self.pipeline.components:
            if c.my_name == component:
                return c

        return None



from typing import Dict, Any

import tensorflow as tf

from kolibri.dnn.layers import L, AttentionWeightedAverageLayer, KMaxPoolingLayer
from kolibri.dnn.tasks.text.classification.base_model import BaseTextClassificationModel
from kolibri.dnn.tasks.text.classification.cnn_attention_model import CNN_Attention_Model
from kolibri.dnn.tasks.text.classification.dpcnn_model import DPCNN_Model


class BiLSTM_Model(BaseTextClassificationModel):

    @classmethod
    def get_default_hyper_parameters(cls):
        return {
            'layer_bi_lstm': {
                'units': 128,
                'return_sequences': False
            },
            'layer_output': {
            }
        }

    def build_model_arc(self):
        output_dim = self.label_indexer.vocab_size

        config = self.hyper_parameters
        embed_model = self.embedding.embed_model

        # build model structure in sequent way
        layer_stack = [
            L.Bidirectional(L.LSTM(**config['layer_bi_lstm'])),
            L.Dense(output_dim, **config['layer_output']),
            self._activation_layer()
        ]

        tensor = embed_model.output
        for layer in layer_stack:
            tensor = layer(tensor)

        self.tf_model = tf.keras.Model(embed_model.inputs, tensor)


class BiGRU_Model(BaseTextClassificationModel):

    @classmethod
    def get_default_hyper_parameters(cls) -> Dict[str, Dict[str, Any]]:
        return {
            'layer_bi_gru': {
                'units': 128,
                'return_sequences': False
            },
            'layer_dense': {
                'activation': 'softmax'
            }
        }

    def build_model_arc(self):
        output_dim = self.label_indexer.vocab_size
        config = self.hyper_parameters
        embed_model = self.embedding.embed_model

        layer_bi_gru = L.Bidirectional(L.GRU(**config['layer_bi_gru']))
        layer_dense = L.Dense(output_dim, **config['layer_dense'])

        tensor = layer_bi_gru(embed_model.output)
        output_tensor = layer_dense(tensor)

        self.tf_model = tf.keras.Model(embed_model.inputs, output_tensor)


class CNN_Model(BaseTextClassificationModel):

    @classmethod
    def get_default_hyper_parameters(cls) -> Dict[str, Dict[str, Any]]:
        return {
            'conv1d_layer': {
                'filters': 128,
                'kernel_size': 5,
                'activation': 'relu'
            },
            'max_pool_layer': {},
            'dense_layer': {
                'units': 64,
                'activation': 'relu'
            },
            'activation_layer': {
                'activation': 'softmax'
            },
            'layer_output': {

            },
            'multipooling': False,
            'dropout': 0.2
        }

    def build_model_arc_(self):

        output_dim = self.label_indexer.vocab_size
        config = self.hyper_parameters

        filters = config['conv1d_layer']['filters']
        filter_sizes = config['conv1d_layer']['kernel_size']

        embed_model = self.embedding.embed_model

        # build model structure in sequent way
        layers_seq = []

        if len(filter_sizes) > 1:
            graph_in = L.Input(shape=(self.sequence_length, self.embedding.embedding_size))

            convolutions = []
            for filter_size in filter_sizes:
                x = L.Conv1D(
                    filters,
                    filter_size,
                    padding='same',
                    activation='relu')(graph_in)
                if config['multipooling']:
                    x = L.MaxPooling1D(filter_size)(x)
                convolutions.append(x)

            graph_out = L.concatenate(convolutions, axis=1)
            graph_out = L.GlobalMaxPooling1D()(graph_out)

            layers_seq.append(tf.keras.Model(graph_in, graph_out))
        else:
            layers_seq.append(L.Conv1D(**config['conv1d_layer']))
        #        layers_seq.append(L.GlobalMaxPooling1D(**config['max_pool_layer']))
        if config['dropout'] > 0:
            layers_seq.append(L.Dropout(config['dropout']))
            layers_seq.append(L.GlobalMaxPooling1D())

        layers_seq.append(L.Dense(**config['dense_layer']))
        layers_seq.append(L.Dense(output_dim, **config['activation_layer']))

        tensor = embed_model.output
        for layer in layers_seq:
            tensor = layer(tensor)

        self.tf_model = tf.keras.Model(embed_model.inputs, tensor)

    def build_model_arc(self) -> None:
        output_dim = self.label_indexer.vocab_size

        config = self.hyper_parameters
        embed_model = self.embedding.embed_model

        # build model structure in sequent way
        layer_stack = [
            L.Conv1D(**config['conv1d_layer']),
            L.GlobalMaxPooling1D(**config['max_pool_layer']),
            L.Dense(**config['dense_layer']),
            L.Dense(output_dim, **config['layer_output']),
            self._activation_layer()
        ]

        tensor = embed_model.output
        for layer in layer_stack:
            tensor = layer(tensor)

        self.tf_model = tf.keras.Model(embed_model.inputs, tensor)


class CNN_LSTM_Model(BaseTextClassificationModel):

    @classmethod
    def get_default_hyper_parameters(cls) -> Dict[str, Dict[str, Any]]:
        return {
            'conv_layer': {
                'filters': 32,
                'kernel_size': 3,
                'padding': 'same',
                'activation': 'relu'
            },
            'max_pool_layer': {
                'pool_size': 2
            },
            'lstm_layer': {
                'units': 100
            },
            'activation_layer': {
                'activation': 'softmax'
            },
        }

    def build_model_arc(self):
        output_dim = self.label_indexer.vocab_size
        config = self.hyper_parameters
        embed_model = self.embedding.embed_model

        layers_seq = []
        layers_seq.append(L.Conv1D(**config['conv_layer']))
        layers_seq.append(L.MaxPooling1D(**config['max_pool_layer']))
        layers_seq.append(L.LSTM(**config['lstm_layer']))
        layers_seq.append(L.Dense(output_dim, **config['activation_layer']))

        tensor = embed_model.output
        for layer in layers_seq:
            tensor = layer(tensor)

        self.tf_model = tf.keras.Model(embed_model.inputs, tensor)


class CNN_GRU_Model(BaseTextClassificationModel):

    @classmethod
    def get_default_hyper_parameters(cls) -> Dict[str, Dict[str, Any]]:
        return {
            'conv_layer': {
                'filters': 32,
                'kernel_size': 3,
                'padding': 'same',
                'activation': 'relu'
            },
            'max_pool_layer': {
                'pool_size': 2
            },
            'gru_layer': {
                'units': 100
            },
            'activation_layer': {
                'activation': 'softmax'
            },
        }

    def build_model_arc(self):
        output_dim = self.label_indexer.vocab_size
        config = self.hyper_parameters
        embed_model = self.embedding.embed_model
        layers_seq = []
        layers_seq.append(L.Conv1D(**config['conv_layer']))
        layers_seq.append(L.MaxPooling1D(**config['max_pool_layer']))
        layers_seq.append(L.GRU(**config['gru_layer']))
        layers_seq.append(L.Dense(output_dim, **config['activation_layer']))

        tensor = embed_model.output
        for layer in layers_seq:
            tensor = layer(tensor)

        self.tf_model = tf.keras.Model(embed_model.inputs, tensor)


class AVCNN_Model(BaseTextClassificationModel):

    @classmethod
    def get_default_hyper_parameters(cls) -> Dict[str, Dict[str, Any]]:
        return {
            'spatial_dropout': {
                'rate': 0.25
            },
            'conv_0': {
                'filters': 300,
                'kernel_size': 1,
                'kernel_initializer': 'normal',
                'padding': 'valid',
                'activation': 'relu'
            },
            'conv_1': {
                'filters': 300,
                'kernel_size': 2,
                'kernel_initializer': 'normal',
                'padding': 'valid',
                'activation': 'relu'
            },
            'conv_2': {
                'filters': 300,
                'kernel_size': 3,
                'kernel_initializer': 'normal',
                'padding': 'valid',
                'activation': 'relu'
            },
            'conv_3': {
                'filters': 300,
                'kernel_size': 4,
                'kernel_initializer': 'normal',
                'padding': 'valid',
                'activation': 'relu'
            },
            # ---
            'attn_0': {},
            'avg_0': {},
            'maxpool_0': {},
            # ---
            'maxpool_1': {},
            'attn_1': {},
            'avg_1': {},
            # ---
            'maxpool_2': {},
            'attn_2': {},
            'avg_2': {},
            # ---
            'maxpool_3': {},
            'attn_3': {},
            'avg_3': {},
            # ---
            'v_col3': {
                # 'mode': 'concat',
                'axis': 1
            },
            'merged_tensor': {
                # 'mode': 'concat',
                'axis': 1
            },
            'dropout': {
                'rate': 0.7
            },
            'dense': {
                'units': 144,
                'activation': 'relu'
            },
            'activation_layer': {
                'activation': 'softmax'
            },
        }

    def build_model_arc(self):
        output_dim = self.label_indexer.vocab_size
        config = self.hyper_parameters
        embed_model = self.embedding.embed_model

        layer_embed_dropout = L.SpatialDropout1D(**config['spatial_dropout'])
        layers_conv = [L.Conv1D(**config[f'conv_{i}']) for i in range(4)]
        layers_sensor = []
        layers_sensor.append(L.GlobalMaxPooling1D())
        layers_sensor.append(AttentionWeightedAverageLayer())
        layers_sensor.append(L.GlobalAveragePooling1D())
        layer_view = L.Concatenate(**config['v_col3'])
        layer_allviews = L.Concatenate(**config['merged_tensor'])
        layers_seq = []
        layers_seq.append(L.Dropout(**config['dropout']))
        layers_seq.append(L.Dense(**config['dense']))
        layers_seq.append(L.Dense(output_dim, **config['activation_layer']))

        embed_tensor = layer_embed_dropout(embed_model.output)
        tensors_conv = [layer_conv(embed_tensor) for layer_conv in layers_conv]
        tensors_matrix_sensor = []
        for tensor_conv in tensors_conv:
            tensor_sensors = []
            tensor_sensors = [layer_sensor(tensor_conv) for layer_sensor in layers_sensor]
            # tensor_sensors.append(L.GlobalMaxPooling1D()(tensor_conv))
            # tensor_sensors.append(AttentionWeightedAverageLayer()(tensor_conv))
            # tensor_sensors.append(L.GlobalAveragePooling1D()(tensor_conv))
            tensors_matrix_sensor.append(tensor_sensors)
        tensors_views = [layer_view(list(tensors)) for tensors in zip(*tensors_matrix_sensor)]
        tensor = layer_allviews(tensors_views)
        # tensors_v_cols = [L.concatenate(tensors, **config['v_col3']) for tensors
        #                   in zip(*tensors_matrix_sensor)]
        # tensor = L.concatenate(tensors_v_cols, **config['merged_tensor'])
        for layer in layers_seq:
            tensor = layer(tensor)

        self.tf_model = tf.keras.Model(embed_model.inputs, tensor)


class KMax_CNN_Model(BaseTextClassificationModel):

    @classmethod
    def get_default_hyper_parameters(cls) -> Dict[str, Dict[str, Any]]:
        return {
            'spatial_dropout': {
                'rate': 0.2
            },
            'conv_0': {
                'filters': 180,
                'kernel_size': 1,
                'kernel_initializer': 'normal',
                'padding': 'valid',
                'activation': 'relu'
            },
            'conv_1': {
                'filters': 180,
                'kernel_size': 2,
                'kernel_initializer': 'normal',
                'padding': 'valid',
                'activation': 'relu'
            },
            'conv_2': {
                'filters': 180,
                'kernel_size': 3,
                'kernel_initializer': 'normal',
                'padding': 'valid',
                'activation': 'relu'
            },
            'conv_3': {
                'filters': 180,
                'kernel_size': 4,
                'kernel_initializer': 'normal',
                'padding': 'valid',
                'activation': 'relu'
            },
            'maxpool_i4': {
                'k': 3
            },
            'merged_tensor': {
                # 'mode': 'concat',
                'axis': 1
            },
            'dropout': {
                'rate': 0.6
            },
            'dense': {
                'units': 144,
                'activation': 'relu'
            },
            'activation_layer': {
                'activation': 'softmax'
            },
        }

    def build_model_arc(self):
        output_dim = self.label_indexer.vocab_size
        config = self.hyper_parameters
        embed_model = self.embedding.embed_model

        layer_embed_dropout = L.SpatialDropout1D(**config['spatial_dropout'])
        layers_conv = [L.Conv1D(**config[f'conv_{i}']) for i in range(4)]
        layers_sensor = [KMaxPoolingLayer(**config['maxpool_i4']),
                         L.Flatten()]
        layer_concat = L.Concatenate(**config['merged_tensor'])
        layers_seq = []
        layers_seq.append(L.Dropout(**config['dropout']))
        layers_seq.append(L.Dense(**config['dense']))
        layers_seq.append(L.Dense(output_dim, **config['activation_layer']))

        embed_tensor = layer_embed_dropout(embed_model.output)
        tensors_conv = [layer_conv(embed_tensor) for layer_conv in layers_conv]
        tensors_sensor = []
        for tensor_conv in tensors_conv:
            tensor_sensor = tensor_conv
            for layer_sensor in layers_sensor:
                tensor_sensor = layer_sensor(tensor_sensor)
            tensors_sensor.append(tensor_sensor)
        tensor = layer_concat(tensors_sensor)
        # tensor = L.concatenate(tensors_sensor, **config['merged_tensor'])
        for layer in layers_seq:
            tensor = layer(tensor)

        self.tf_model = tf.keras.Model(embed_model.inputs, tensor)


class R_CNN_Model(BaseTextClassificationModel):

    @classmethod
    def get_default_hyper_parameters(cls) -> Dict[str, Dict[str, Any]]:
        return {
            'spatial_dropout': {
                'rate': 0.2
            },
            'rnn_0': {
                'units': 64,
                'return_sequences': True
            },
            'conv_0': {
                'filters': 128,
                'kernel_size': 2,
                'kernel_initializer': 'normal',
                'padding': 'valid',
                'activation': 'relu',
                'strides': 1
            },
            'maxpool': {},
            'attn': {},
            'average': {},
            'concat': {
                'axis': 1
            },
            'dropout': {
                'rate': 0.5
            },
            'dense': {
                'units': 120,
                'activation': 'relu'
            },
            'activation_layer': {
                'activation': 'softmax'
            },
        }

    def build_model_arc(self):
        output_dim = self.label_indexer.vocab_size
        config = self.hyper_parameters
        embed_model = self.embedding.embed_model

        layers_rcnn_seq = []
        layers_rcnn_seq.append(L.SpatialDropout1D(**config['spatial_dropout']))
        layers_rcnn_seq.append(L.Bidirectional(L.GRU(**config['rnn_0'])))
        layers_rcnn_seq.append(L.Conv1D(**config['conv_0']))

        layers_sensor = []
        layers_sensor.append(L.GlobalMaxPooling1D())
        layers_sensor.append(AttentionWeightedAverageLayer())
        layers_sensor.append(L.GlobalAveragePooling1D())
        layer_concat = L.Concatenate(**config['concat'])

        layers_full_connect = []
        layers_full_connect.append(L.Dropout(**config['dropout']))
        layers_full_connect.append(L.Dense(**config['dense']))
        layers_full_connect.append(L.Dense(output_dim, **config['activation_layer']))

        tensor = embed_model.output
        for layer in layers_rcnn_seq:
            tensor = layer(tensor)

        tensors_sensor = [layer(tensor) for layer in layers_sensor]
        tensor_output = layer_concat(tensors_sensor)
        # tensor_output = L.concatenate(tensor_sensors, **config['concat'])

        for layer in layers_full_connect:
            tensor_output = layer(tensor_output)

        self.tf_model = tf.keras.Model(embed_model.inputs, tensor_output)


class AVRNN_Model(BaseTextClassificationModel):

    @classmethod
    def get_default_hyper_parameters(cls) -> Dict[str, Dict[str, Any]]:
        return {
            'spatial_dropout': {
                'rate': 0.25
            },
            'rnn_0': {
                'units': 60,
                'return_sequences': True
            },
            'rnn_1': {
                'units': 60,
                'return_sequences': True
            },
            'concat_rnn': {
                'axis': 2
            },
            'last': {},
            'maxpool': {},
            'attn': {},
            'average': {},
            'all_views': {
                'axis': 1
            },
            'dropout': {
                'rate': 0.5
            },
            'dense': {
                'units': 144,
                'activation': 'relu'
            },
            'activation_layer': {
                'activation': 'softmax'
            },
        }

    def build_model_arc(self):
        output_dim = self.label_indexer.vocab_size
        config = self.hyper_parameters
        embed_model = self.embedding.embed_model

        layers_rnn0 = []
        layers_rnn0.append(L.SpatialDropout1D(**config['spatial_dropout']))
        layers_rnn0.append(L.Bidirectional(L.GRU(**config['rnn_0'])))

        layer_bi_rnn1 = L.Bidirectional(L.GRU(**config['rnn_1']))

        layer_concat = L.Concatenate(**config['concat_rnn'])

        layers_sensor = []
        layers_sensor.append(L.Lambda(lambda t: t[:, -1], name='last'))
        layers_sensor.append(L.GlobalMaxPooling1D())
        layers_sensor.append(AttentionWeightedAverageLayer())
        layers_sensor.append(L.GlobalAveragePooling1D())

        layer_allviews = L.Concatenate(**config['all_views'])
        layers_full_connect = []
        layers_full_connect.append(L.Dropout(**config['dropout']))
        layers_full_connect.append(L.Dense(**config['dense']))
        layers_full_connect.append(L.Dense(output_dim, **config['activation_layer']))

        tensor_rnn = embed_model.output
        for layer in layers_rnn0:
            tensor_rnn = layer(tensor_rnn)
        tensor_concat = layer_concat([tensor_rnn, layer_bi_rnn1(tensor_rnn)])
        tensor_sensors = [layer(tensor_concat) for layer in layers_sensor]
        tensor_output = layer_allviews(tensor_sensors)
        for layer in layers_full_connect:
            tensor_output = layer(tensor_output)

        self.tf_model = tf.keras.Model(embed_model.inputs, tensor_output)


class Dropout_BiGRU_Model(BaseTextClassificationModel):

    @classmethod
    def get_default_hyper_parameters(cls) -> Dict[str, Dict[str, Any]]:
        return {
            'spatial_dropout': {
                'rate': 0.15
            },
            'rnn_0': {
                'units': 64,
                'return_sequences': True
            },
            'dropout_rnn': {
                'rate': 0.35
            },
            'rnn_1': {
                'units': 64,
                'return_sequences': True
            },
            'last': {},
            'maxpool': {},
            'average': {},
            'all_views': {
                'axis': 1
            },
            'dropout': {
                'rate': 0.5
            },
            'dense': {
                'units': 72,
                'activation': 'relu'
            },
            'activation_layer': {
                'activation': 'softmax'
            },
        }

    def build_model_arc(self):
        output_dim = self.label_indexer.vocab_size
        config = self.hyper_parameters
        embed_model = self.embedding.embed_model

        layers_rnn = []
        layers_rnn.append(L.SpatialDropout1D(**config['spatial_dropout']))
        layers_rnn.append(L.Bidirectional(L.GRU(**config['rnn_0'])))
        layers_rnn.append(L.Dropout(**config['dropout_rnn']))
        layers_rnn.append(L.Bidirectional(L.GRU(**config['rnn_1'])))

        layers_sensor = []
        layers_sensor.append(L.Lambda(lambda t: t[:, -1], name='last'))
        layers_sensor.append(L.GlobalMaxPooling1D())
        layers_sensor.append(L.GlobalAveragePooling1D())

        layer_allviews = L.Concatenate(**config['all_views'])

        layers_full_connect = []
        layers_full_connect.append(L.Dropout(**config['dropout']))
        layers_full_connect.append(L.Dense(**config['dense']))
        layers_full_connect.append(L.Dense(output_dim, **config['activation_layer']))

        tensor_rnn = embed_model.output
        for layer in layers_rnn:
            tensor_rnn = layer(tensor_rnn)
        tensor_sensors = [layer(tensor_rnn) for layer in layers_sensor]
        tensor_output = layer_allviews(tensor_sensors)
        for layer in layers_full_connect:
            tensor_output = layer(tensor_output)

        self.tf_model = tf.keras.Model(embed_model.inputs, tensor_output)


class Dropout_AVRNN_Model(BaseTextClassificationModel):

    @classmethod
    def get_default_hyper_parameters(cls) -> Dict[str, Dict[str, Any]]:
        return {
            'spatial_dropout': {
                'rate': 0.25
            },
            'rnn_0': {
                'units': 56,
                'return_sequences': True
            },
            'rnn_dropout': {
                'rate': 0.3
            },
            'rnn_1': {
                'units': 56,
                'return_sequences': True
            },
            'last': {},
            'maxpool': {},
            'attn': {},
            'average': {},
            'all_views': {
                'axis': 1
            },
            'dropout_0': {
                'rate': 0.5
            },
            'dense': {
                'units': 128,
                'activation': 'relu'
            },
            'dropout_1': {
                'rate': 0.25
            },
            'activation_layer': {
                'activation': 'softmax'
            },
        }

    def build_model_arc(self):
        output_dim = self.label_indexer.vocab_size
        config = self.hyper_parameters
        embed_model = self.embedding.embed_model

        layers_rnn = []
        layers_rnn.append(L.SpatialDropout1D(**config['spatial_dropout']))
        layers_rnn.append(L.Bidirectional(L.GRU(**config['rnn_0'])))
        layers_rnn.append(L.SpatialDropout1D(**config['rnn_dropout']))
        layers_rnn.append(L.Bidirectional(L.GRU(**config['rnn_1'])))

        layers_sensor = []
        layers_sensor.append(L.Lambda(lambda t: t[:, -1], name='last'))
        layers_sensor.append(L.GlobalMaxPooling1D())
        layers_sensor.append(AttentionWeightedAverageLayer())
        layers_sensor.append(L.GlobalAveragePooling1D())

        layer_allviews = L.Concatenate(**config['all_views'])
        layers_full_connect = []
        layers_full_connect.append(L.Dropout(**config['dropout_0']))
        layers_full_connect.append(L.Dense(**config['dense']))
        layers_full_connect.append(L.Dropout(**config['dropout_1']))
        layers_full_connect.append(L.Dense(output_dim, **config['activation_layer']))

        tensor_rnn = embed_model.output
        for layer in layers_rnn:
            tensor_rnn = layer(tensor_rnn)
        tensor_sensors = [layer(tensor_rnn) for layer in layers_sensor]
        tensor_output = layer_allviews(tensor_sensors)
        for layer in layers_full_connect:
            tensor_output = layer(tensor_output)

        self.tf_model = tf.keras.Model(embed_model.inputs, tensor_output)


def get_model(model_type, embedding=None, hyper_parameters=None):
    if model_type == 'bilstm':
        return BiLSTM_Model(embedding=embedding, hyper_parameters=hyper_parameters)
    elif model_type == 'bigru':
        return BiGRU_Model(embedding=embedding, hyper_parameters=hyper_parameters)
    elif model_type == 'cnn':
        return CNN_Model(embedding=embedding, hyper_parameters=hyper_parameters)
    elif model_type == 'cnn_lstm':
        return CNN_LSTM_Model(embedding=embedding, hyper_parameters=hyper_parameters)
    elif model_type == 'cnn_gru':
        return CNN_GRU_Model(embedding=embedding, hyper_parameters=hyper_parameters)
    elif model_type == 'avcnn':
        return AVCNN_Model(embedding=embedding, hyper_parameters=hyper_parameters)
    elif model_type == 'kmax_cnn':
        return KMax_CNN_Model(embedding=embedding, hyper_parameters=hyper_parameters)
    elif model_type == 'r_cnn':
        return R_CNN_Model(embedding=embedding, hyper_parameters=hyper_parameters)
    if model_type == 'avrnn':
        return AVRNN_Model(embedding=embedding, hyper_parameters=hyper_parameters)
    elif model_type == 'dropout_bigru':
        return Dropout_BiGRU_Model(embedding=embedding, hyper_parameters=hyper_parameters)
    elif model_type == 'dropout_avrnn':
        return Dropout_AVRNN_Model(embedding=embedding, hyper_parameters=hyper_parameters)
    elif model_type == 'dpcnn':
        return DPCNN_Model(embedding=embedding, hyper_parameters=hyper_parameters)
    elif model_type == 'cnn_attention':
        return CNN_Attention_Model(embedding=embedding, hyper_parameters=hyper_parameters)
    else:
        raise Exception('Model does not exist in the library of text classification models')


if __name__ == "__main__":
    print(BiLSTM_Model.get_default_hyper_parameters())
    # logging.basicConfig(level=logging.DEBUG)
    from kolibri.data.corpus.corpus import ConsumerComplaintsCorpus

    corpus = ConsumerComplaintsCorpus()
    x_train, y_train = corpus.get_data()

    #    from kolibri.dnn.embeddings import DefaultEmbedding

    #    indexer = LabelIndexer(multi_label=False)

    #    embed = DefaultEmbedding()

    m = CNN_Model()

    x_train = [x.split() for x in x_train]
    # m.build_model(x, y)
    m.fit(x_train, y_train, epochs=1)

    print(m.predict(x_train[:10]))
    # m.evaluate(x, y)
#   print(m.predict_top_k_class(x[:10]))



#!/usr/bin/env python

import numpy as np


# Sample-based measures:
# Hamming Loss, Accuracy, Precision, Recall, F1-Score:
def calculate_hamming_loss(real_labels, predicted_labels):
    num_samples, num_classes = real_labels.shape
    xor_arrays = np.logical_xor(real_labels, predicted_labels).astype('float')
    sum_rows = np.sum(xor_arrays, axis=1) / num_classes
    hamming_loss = sum_rows.sum() / num_samples
    return hamming_loss


def calculate_accuracy(real_labels, predicted_labels, average=True):
    num_samples = real_labels.shape[0]
    true_positive = np.sum(np.logical_and(real_labels, predicted_labels), axis=1).astype('float')
    tp_fn_fp = np.sum(np.logical_or(real_labels, predicted_labels), axis=1).astype('float')
    acc_per_sample = true_positive / tp_fn_fp
    accuracy = np.sum(acc_per_sample) / num_samples
    if average:
        return accuracy
    else:
        return acc_per_sample


def calculate_precision(real_labels, predicted_labels, average=True):
    num_samples = real_labels.shape[0]
    true_positive = np.sum(np.logical_and(real_labels, predicted_labels), axis=1).astype('float')
    num_real_labels = np.sum(real_labels, axis=1).astype('float')
    precision_per_sample = true_positive / num_real_labels
    precision = np.sum(precision_per_sample) / num_samples
    if average:
        return precision
    else:
        return precision_per_sample


def calculate_recall(real_labels, predicted_labels, average=True):
    num_samples = real_labels.shape[0]
    true_positive = np.sum(np.logical_and(real_labels, predicted_labels), axis=1).astype('float')
    num_predicted_labels = np.sum(predicted_labels, axis=1).astype('float')
    num_predicted_labels[num_predicted_labels == 0] = np.nan
    recall_per_sample = true_positive / num_predicted_labels
    recall = np.nansum(recall_per_sample) / num_samples
    if average:
        return recall
    else:
        recall_per_sample[np.isnan(recall_per_sample)] = 0
        return recall_per_sample


def calculate_f1_score(real_labels, predicted_labels):
    num_samples = real_labels.shape[0]
    recall = calculate_recall(real_labels, predicted_labels, average=False)
    precision = calculate_precision(real_labels, predicted_labels, average=False)
    f1_num = 2 * recall * precision
    f1_den = recall + precision
    f1_den[f1_den == 0] = np.nan
    f1_score = f1_num / f1_den
    f1_score = np.nansum(f1_score) / num_samples
    return f1_score


# Label-based measures:
# Macro Precision, Recall and F1-Score
def calculate_macro_precision(real_labels, predicted_labels, average=True):
    num_classes = real_labels.shape[1]
    true_positive = np.sum(np.logical_and(real_labels, predicted_labels), axis=0).astype('float')
    false_positive = np.sum(np.logical_and(np.logical_not(real_labels), predicted_labels), axis=0).astype('float')
    precision_den = true_positive + false_positive
    precision_den[precision_den == 0] = np.nan
    macro_precision_per_class = true_positive / precision_den
    macro_precision = np.nansum(macro_precision_per_class) / num_classes
    if average:
        return macro_precision
    else:
        macro_precision_per_class[np.isnan(macro_precision_per_class)] = 0
        return macro_precision_per_class


def calculate_macro_recall(real_labels, predicted_labels, average=True):
    num_classes = real_labels.shape[1]
    true_positive = np.sum(np.logical_and(real_labels, predicted_labels), axis=0).astype('float')
    false_negative = np.sum(np.logical_and(real_labels, np.logical_not(predicted_labels)), axis=0).astype('float')
    recall_den = true_positive + false_negative
    recall_den[recall_den == 0] = np.nan
    macro_recall_per_class = true_positive / recall_den
    macro_recall = np.nansum(macro_recall_per_class) / num_classes
    if average:
        return macro_recall
    else:
        macro_recall_per_class[np.isnan(macro_recall_per_class)] = 0
        return macro_recall_per_class


def calculate_macro_f1_score(real_labels, predicted_labels, per_class=False):
    num_classes = real_labels.shape[1]
    recall = calculate_macro_recall(real_labels, predicted_labels, average=False)
    precision = calculate_macro_precision(real_labels, predicted_labels, average=False)
    f1_num = 2 * recall * precision
    f1_den = recall + precision
    f1_den[f1_den == 0] = np.nan
    f1_score = f1_num / f1_den
    if per_class:
        f1_score[np.isnan(f1_score)] = 0
        return f1_score
    else:
        f1_score = np.nansum(f1_score) / num_classes
        return f1_score


# Overall measures:
# Micro F1-Score, Precision and Recall
def calculate_micro_precision(real_labels, predicted_labels):
    true_positive = np.sum(np.logical_and(real_labels, predicted_labels)).astype('float')
    false_positive = np.sum(np.logical_and(np.logical_not(real_labels), predicted_labels)).astype('float')
    precision_den = true_positive + false_positive
    if precision_den == 0:
        precision = 0
    else:
        precision = true_positive / precision_den
    return precision


def calculate_micro_recall(real_labels, predicted_labels):
    true_positive = np.sum(np.logical_and(real_labels, predicted_labels)).astype('float')
    false_negative = np.sum(np.logical_and(real_labels, np.logical_not(predicted_labels))).astype('float')
    recall_den = true_positive + false_negative
    if recall_den == 0:
        recall = 0
    else:
        recall = true_positive / recall_den
    return recall


def calculate_micro_f1_score(real_labels, predicted_labels):
    precision = calculate_micro_precision(real_labels, predicted_labels)
    recall = calculate_micro_recall(real_labels, predicted_labels)
    f1_num = 2 * precision * recall
    f1_den = precision + recall
    if f1_den == 0:
        f1_score = 0
    else:
        f1_score = f1_num / f1_den
    return f1_score


def calculate_micro_f1_score_from_precision_recall(precision, recall):
    f1_num = 2 * precision * recall
    f1_den = precision + recall
    if f1_den == 0:
        f1_score = 0
    else:
        f1_score = f1_num / f1_den
    return f1_score



import os
import tempfile
import time
import unittest

from kolibri.data.corpus.corpus import ConsumerComplaintsCorpus
from kolibri.dnn.embeddings import WordEmbedding
from kolibri.dnn.tasks.text.classification import BiLSTM_Model
from kolibri.tokenizer import WordTokenizer
from tests.dnn.text.test_embeddings.test_word_embedding import sample_w2v_path


class TestBiLSTM_Model(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.EPOCH_COUNT = 2
        cls.TASK_MODEL_CLASS = BiLSTM_Model
        cls.w2v_embedding = WordEmbedding(sample_w2v_path)

    @classmethod
    def tearDownClass(cls) -> None:
        cls.w2v_embedding = None

    def test_basic_use(self):
        tokenizer = WordTokenizer()
        model = self.TASK_MODEL_CLASS(sequence_length=20)
        corpus = ConsumerComplaintsCorpus()
        train_x, train_y = corpus.get_data(nb_samples=500)
        valid_x, valid_y = train_x, train_y
        train_x = tokenizer.transform(train_x)
        valid_x = tokenizer.transform(valid_x)
        model.fit(train_x,
                  train_y,
                  x_validate=valid_x,
                  y_validate=valid_y,
                  epochs=self.EPOCH_COUNT)

        model_path = os.path.join(tempfile.gettempdir(), str(time.time()))
        original_y = model.predict(train_x[:20])
        model.save(model_path)

        # Make sure use sigmoid as activation function
        assert model.tf_model.layers[-1].activation.__name__ == 'softmax'

        del model
        new_model = self.TASK_MODEL_CLASS.load_model(model_path)
        new_model.tf_model.summary()
        new_y = new_model.predict(train_x[:20])
        assert new_y == original_y

        report = new_model.evaluate(valid_x, valid_y)
        for key in ['precision', 'recall', 'f1-score', 'support', 'detail']:
            assert key in report

        # Make sure use sigmoid as activation function
        assert new_model.tf_model.layers[-1].activation.__name__ == 'softmax'

    # def test_multi_label(self):
    #     corpus=ConsumerComplaintsCorpus()
    #     model = self.TASK_MODEL_CLASS(sequence_length=20, multi_label=True)
    #     x, y = corpus.get_data(nb_samples=500)
    #     model.fit(x, y, epochs=self.EPOCH_COUNT)
    #
    #     model_path = os.path.join(tempfile.gettempdir(), str(time.time()))
    #     original_y = model.predict(x[:20])
    #     model.save(model_path)
    #
    #     # Make sure use sigmoid as activation function
    #     assert model.tf_model.layers[-1].activation.__name__ == 'sigmoid'
    #     del model
    #
    #     new_model = self.TASK_MODEL_CLASS.load_model(model_path)
    #     new_model.tf_model.summary()
    #     new_y = new_model.predict(x[:20])
    #
    #     assert new_y == original_y
    #
    #     report = new_model.evaluate(x, y)
    #     for key in ['precision', 'recall', 'f1-score', 'support', 'detail']:
    #         assert key in report
    #
    #     # Make sure use sigmoid as activation function
    #     assert new_model.tf_model.layers[-1].activation.__name__ == 'sigmoid'

    def test_with_word_embedding(self):
        tokenizer = WordTokenizer()
        model = self.TASK_MODEL_CLASS(embedding=self.w2v_embedding)
        corpus = ConsumerComplaintsCorpus()
        train_x, train_y = corpus.get_data(nb_samples=500)
        valid_x, valid_y = train_x, train_y
        train_x = tokenizer.transform(train_x)
        valid_x = tokenizer.transform(valid_x)

        model.fit(train_x,
                  train_y,
                  x_validate=valid_x,
                  y_validate=valid_y,
                  epochs=self.EPOCH_COUNT)

        model_path = os.path.join(tempfile.gettempdir(), str(time.time()))
        _ = model.predict(valid_x[:20])
        model.save(model_path)

        del model

        new_model = self.TASK_MODEL_CLASS.load_model(model_path)
        new_model.tf_model.summary()
        _ = new_model.predict(valid_x[:20])


if __name__ == '__main__':
    unittest.main()



import numpy as np
import tensorflow as tf
from scipy.sparse import coo_matrix
from tensorflow.keras import backend as K


def confusion_matrix(y_true, y_pred, sample_weight=None):
    y_pred = np.array(y_pred)
    y_true = np.array(y_true)
    labels = np.unique(np.hstack([y_true, y_pred]))
    n_labels = len(labels)

    if sample_weight is None:
        sample_weight = np.ones(y_true.shape[0], dtype=np.int64)
    else:
        sample_weight = np.asarray(sample_weight)

    #    n_labels = y_values.size
    #    label_to_ind = {y: x for x, y in enumerate(y_values)}
    # convert yt, yp into index
    #    y_pred = np.array([label_to_ind.get(x, n_labels + 1) for x in y_pred])
    #    y_true = np.array([label_to_ind.get(x, n_labels + 1) for x in y_true])

    # intersect y_pred, y_true with y_values, eliminate items not in y_values
    #    ind = np.logical_and(y_pred < n_labels, y_true < n_labels)
    #    y_pred = y_pred[ind]
    #    y_true = y_true[ind]
    # also eliminate weights of eliminated items
    #    sample_weight = sample_weight[ind]

    # Choose the accumulator dtype to always have high precision

    cm = coo_matrix((sample_weight, (y_true, y_pred)),
                    shape=(n_labels, n_labels),
                    ).toarray()

    cm = np.nan_to_num(cm)

    return cm


def mean_false_error(y_true, y_pred):
    diff = 0.5 * (K.square(y_true - y_pred))
    segments = K.argmax(y_true)

    return K.sum(K.square(tf.math.unsorted_segment_mean(
        diff, segments, num_segments=K.max(segments) + 1
    )), axis=0)


def mean_false_error_2(y_true, y_pred):
    neg_y_true = 1 - y_true
    neg_y_pred = 1 - y_pred

    FN = K.sum(y_true * neg_y_pred, axis=0)
    FP = K.sum(neg_y_true * y_pred, axis=0)
    TP = K.sum(y_true * y_pred, axis=0)
    TN = K.sum(neg_y_true * neg_y_pred, axis=0)

    fpe = FP / (TP + FP)
    fne = FN / (TN + FN)
    score = fne * fne + fpe * fpe
    score = tf.where(tf.math.is_nan(score), tf.ones_like(score) * 0, score)
    return score


def mean_squared_false_error(y_true, y_preficted):
    m = confusion_matrix(y_true, y_preficted)
    class_counts = m.sum(axis=1)
    FP = m.sum(axis=0) - np.diag(m)
    FN = m.sum(axis=1) - np.diag(m)

    fpe = np.mean(FP / class_counts)
    fne = np.mean(FN / class_counts)
    return (fpe * fpe) + (fne * fne)


def mean_false_error2(y_true, y_pred):
    diff = 0.5 * (K.square(y_true - y_pred))
    segments = K.argmax(y_true)

    return K.sum(tf.math.unsorted_segment_mean(
        diff, segments, num_segments=K.max(segments) + 1
    ), axis=0)


if __name__ == '__main__':
    y_true = K.variable([[0, 0, 1], [1, 0, 0], [0, 0, 1], [0, 0, 1], [1, 0, 0], [0, 1, 0]])
    y_pred = K.variable([[1, 0, 0], [1, 0, 0], [0, 0, 1], [0, 0, 1], [1, 0, 0], [0, 0, 1]])
    print(K.eval(mean_false_error2(y_true, y_pred)))
#  y_true = np.array([1,0, 1, 1,0 ,0])
#  y_pred = np.array([0,0, 1,1,0, 1])
# #
#  print(mean_false_error(y_true, y_pred))
#



import matplotlib.pyplot as plt
import numpy as np
from matplotlib import cm

from kolibri.classifier_evaluator.metrics import roc, roc_auc
from kolibri.classifier_evaluator.pre_process import data_type_converter

# setting standard plot size:
_DEFAULT_FIGURE_SIZE = (10, 10)
_DEFAULT_COLOR_PALETTE = "jet"


def plot_roc(roc_dict: dict, **kwargs):
    """
    plot roc curves;
    :param roc_dict: dict, that contains series needed for the plot; i.e.,
        set 1: given fpr and tpr:
            {model_name: {"fpr": [pd.Series, np.ndarray, list], "tpr": [pd.Series, np.ndarray, list], ...}}
        set 2: given true series and predicted_prob series:
            {model_name: {"true": [pd.Series, np.ndarray, list], "predicted_prob": [pd.Series, np.ndarray, list], ...}}
    :param kwargs:
        :param figure_size: tuple, e.g., (10, 10), the size of the figure;
        :param subtitle: str, sub title given to the plot;
        :param save: [bool, str], if it is saved to an external .png file;
    :return: plt.show or saved image;
    """
    roc_element_dict = dict()
    for model_name, series_dict in roc_dict.items():
        if ('fpr' in series_dict) and ('tpr' in series_dict):
            roc_element_dict[model_name] = {'fpr': data_type_converter(series_dict['fpr']),
                                            'tpr': data_type_converter(series_dict['tpr'])}
        elif ('true' in series_dict) and ('predicted_prob' in series_dict):
            if 'pos_label' in series_dict:
                fpr, tpr, thresholds = roc(true=data_type_converter(series_dict['true']),
                                           predicted_prob=data_type_converter(series_dict['predicted_prob']),
                                           pos_label=series_dict['pos_label'])
            else:
                fpr, tpr, thresholds = roc(true=data_type_converter(series_dict['true']),
                                           predicted_prob=data_type_converter(series_dict['predicted_prob']))

            roc_element_dict[model_name] = {'fpr': fpr,
                                            'tpr': tpr}
        else:
            raise KeyError(f"series dict does not contain needed series ({list(series_dict.keys())})")

    # compute auc if not present:
    for model_name, series_dict in roc_element_dict.items():
        if 'auc' in roc_dict[model_name]:
            series_dict['auc'] = roc_dict[model_name]['auc']
        else:
            series_dict['auc'] = roc_auc(**series_dict)

    # plot set up:
    if 'figure_size' in kwargs:
        figure_size = kwargs['figure_size']
    else:
        figure_size = _DEFAULT_FIGURE_SIZE

    fig, ax = plt.subplots(figsize=figure_size)
    n_model = len(roc_element_dict)
    cm_subsection = np.linspace(0, 1, n_model)
    cmap = cm.get_cmap(_DEFAULT_COLOR_PALETTE)
    colors = [cmap(x) for x in cm_subsection]

    # setting label size:
    axis_label_size = int(min(figure_size) * 1.6)
    title_label_size = int(min(figure_size) * 2.1)
    line_width = int(min(figure_size) / 6)

    ax.set_xlabel('False Positive Rate', fontsize=axis_label_size)
    ax.set_ylabel('True Positive Rate', fontsize=axis_label_size)

    if 'subtitle' in kwargs:
        title_text = f"Reciver Operating Characteristic Plot ({kwargs['subtitle']})"
        saved_file_name = f"roc_{kwargs['subtitle']}.png"
    else:
        title_text = f"Reciver Operating Characteristic Plot"
        saved_file_name = f"roc.png"

    ax.title.set_text(title_text)
    ax.title.set_fontsize(title_label_size)
    plt.xlim([0.0, 1.0])
    plt.ylim([0.0, 1.01])
    ax.grid(True)

    _index = -1
    for model_name, series_dict in roc_element_dict.items():
        _index += 1
        ax.plot(series_dict['fpr'], series_dict['tpr'],
                lw=line_width, alpha=1,
                label=f"Model {model_name} (AUC = {str(round(series_dict['auc'], 2))})",
                color=colors[_index])

    ax.plot([0, 1], [0, 1], linestyle='--', lw=line_width, color='navy', label='Luck', alpha=.8)
    ax.legend(loc="lower right")

    if 'save' in kwargs:
        save = kwargs['save']
        if isinstance(save, str):
            plt.savefig(save)
        else:
            plt.savefig(saved_file_name)

    plt.show()



# encoding: utf-8

# author: BrikerMan
# contact: eliyar917@gmail.com
# blog: https://eliyar.biz

# file: experimental.py
# time: 2019-05-22 19:35

from typing import Dict, Any

from keras_self_attention import SeqSelfAttention
from tensorflow import keras

import kolibri.dnn
from kolibri.dnn.layers import L
from kolibri.dnn.tasks.text.labeling import BaseLabelingModel


class BLSTMAttentionModel(BaseLabelingModel):
    """Bidirectional LSTM Self Attention Sequence Labeling Model"""

    @classmethod
    def get_default_hyper_parameters(cls) -> Dict[str, Dict[str, Any]]:
        """
        Get hyper parameters of model
        Returns:
            hyper parameters dict
        """
        return {
            'layer_blstm': {
                'units': 64,
                'return_sequences': True
            },
            'layer_self_attention': {
                'attention_activation': 'sigmoid'
            },
            'layer_dropout': {
                'rate': 0.5
            },
            'layer_time_distributed': {},
            'layer_activation': {
                'activation': 'softmax'
            }
        }

    def build_model_arc(self):
        """
        build model architectural
        """
        output_dim = len(self.indexer.label2idx)
        config = self.hyper_parameters
        embed_model = self.embedding.embed_model

        layer_blstm = L.Bidirectional(L.LSTM(**config['layer_blstm']),
                                      name='layer_blstm')
        layer_self_attention = SeqSelfAttention(**config['layer_self_attention'],
                                                name='layer_self_attention')
        layer_dropout = L.Dropout(**config['layer_dropout'],
                                  name='layer_dropout')

        layer_time_distributed = L.TimeDistributed(L.Dense(output_dim,
                                                           **config['layer_time_distributed']),
                                                   name='layer_time_distributed')
        layer_activation = L.Activation(**config['layer_activation'])

        tensor = layer_blstm(embed_model.output)
        tensor = layer_self_attention(tensor)
        tensor = layer_dropout(tensor)
        tensor = layer_time_distributed(tensor)
        output_tensor = layer_activation(tensor)

        self.tf_model = keras.Model(embed_model.inputs, output_tensor)


# Register custom layer
kolibri.dnn.custom_objects['SeqSelfAttention'] = SeqSelfAttention

if __name__ == "__main__":
    print("Hello world")



import string
import sys
import warnings

from six.moves import range

from kolibri.stopwords import get_stop_words
from kolibri.tokenizer.tokenizer import *

if sys.version_info < (3,):
    maketrans = string.maketrans
else:
    maketrans = str.maketrans

punctuations = '!"#$%&()*+,-./:;<=>?@[\\]^_`{|}~\t\n'


def remove_punctuations(text, split=' ', filters='!"#$%&()*+,-./:;<=>?@[\\]^_`{|}~\t\n'):
    if sys.version_info < (3,):
        if isinstance(text, unicode):  # noqa: F821
            translate_map = {
                ord(c): unicode(split) for c in filters  # noqa: F821
            }
            text = text.translate(translate_map)
        elif len(split) == 1:
            translate_map = maketrans(filters, split * len(filters))
            text = text.translate(translate_map)
        else:
            for c in filters:
                text = text.replace(c, split)
    else:
        translate_dict = {c: split for c in filters}
        translate_map = maketrans(translate_dict)
        text = text.translate(translate_map)
    return text.strip()


def ngram(text, n=1, stride=None):
    if stride == None:
        stride = n
    return [text[i:i + n] for i in range(0, len(text), stride)]


def text_to_word_sequence(text,
                          filters=None,
                          lower=True, split=" "):
    """Converts a text to a sequence of words (or tokens).
    # Arguments
        text: Input text (string).
        filters: list (or concatenation) of characters to filter out, such as
            punctuation. Default: ``!"#$%&()*+,-./:;<=>?@[\\]^_`{|}~\\t\\n``,
            includes basic punctuation, tabs, and newlines.
        lower: boolean. Whether to convert the input to lowercase.
        split: str. Separator for word splitting.
    # Returns
        A list of words (or tokens).
    """
    if lower:
        text = text.lower()

    if filters:
        text = remove_punctuations(text, split, filters)
    seq = text.split(split)
    return [i for i in seq if i]


class CharTokenizer(Tokenizer):
    """Text tokenization utility class inspired from keras
    """
    name = "char_tokenizer"

    provides = ["tokens"]
    defaults = {
        "remove-stopwros": False,
        "language": 'en',
        "lower-case": True,
        "remove-punct": False,
        "ngram": 1
    }

    def __init__(self, config, **kwargs):
        # Legacy support
        super().__init__(config)
        if 'nb_words' in kwargs:
            warnings.warn('The `nb_words` argument in `Tokenizer` '
                          'has been renamed `num_words`.')
            num_words = kwargs.pop('nb_words')
        if kwargs:
            raise TypeError('Unrecognized keyword arguments: ' + str(kwargs))
        self.lower = self.component_config['lower-case']
        self.remove_stopwords = self.component_config['remove-stopwros']
        self.remove_punctuations = self.component_config["remove-punct"]
        self.punctuations = None
        if self.remove_punctuations:
            self.punctuations = punctuations
        if self.remove_stopwords:
            self.stopwords = get_stop_words(self.component_config['language'])
        self.ngram = self.component_config['ngram']

    def fit(self, training_data, target):
        return self

    def transform(self, texts, **kwargs):
        return self.texts_to_sequences(texts)

    def tokenize(self, text):
        return self.texts_to_sequences([text])

    def process(self, document, **kwargs):
        if hasattr(document, '_sentences'):
            document.tokens = []
            for sentence in document.sentences:
                document.tokens.append(self.texts_to_sequences(sentence))
            document.sentences = None
        else:
            document.tokens = self.texts_to_sequences([document.text])[0]
            document.raw_text = None

    def get_info(self):
        return "char_tokenizer"

    def texts_to_sequences(self, texts):
        """Transforms each text in texts to a sequence of integers.
        Only top `num_words-1` most frequent words will be taken into account.
        Only words known by the tokenizer will be taken into account.
        # Arguments
            texts: A list of texts (strings).
        # Returns
            A list of sequences.
        """

        return list(self.texts_to_sequences_generator(texts))

    def texts_to_sequences_generator(self, texts):
        """Transforms each text in `texts` to a sequence of integers.
        Each item in texts can also be a list,
        in which case we assume each item of that list to be a token.
        Only top `num_words-1` most frequent words will be taken into account.
        Only words known by the tokenizer will be taken into account.
        # Arguments
            texts: A list of texts (strings).
        # Yields
            Yields individual sequences.
        """

        for text in texts:
            if self.remove_stopwords:
                text = text_to_word_sequence(text, filters=self.punctuations)
                text = ' '.join([t for t in text if t not in self.stopwords])
            elif self.remove_punctuations:
                text = remove_punctuations(text, filters=self.punctuations)

            if self.lower:
                text = text.lower()
            yield ngram(text, self.ngram)



import re

from kolibri.application.datastore import DatabaseAdapter


class MongoDatabaseAdapter(DatabaseAdapter):
    """
    The MongoDatabaseAdapter is an interface that allows
    kolibri to store documents in a MongoDB database.

    :keyword database_uri: The URI of a remote instance of MongoDB.
                           This can be any valid
                           `MongoDB connection string <https://docs.mongodb.com/manual/reference/connection-string/>`_
    :type database_uri: str

    .. code-block:: python

       database_uri='mongodb://example.com:8100/'
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        from pymongo import MongoClient
        from pymongo.errors import OperationFailure

        self.database_uri = kwargs.get(
            'database_uri', 'mongodb://localhost:27017/kolibri-database'
        )

        # Use the default host and port
        self.client = MongoClient(self.database_uri)

        # Increase the sort buffer to 42M if possible
        try:
            self.client.admin.command({'setParameter': 1, 'internalQueryExecMaxBlockingSortBytes': 44040192})
        except OperationFailure:
            pass

        # Specify the name of the database
        self.database = self.client.get_database()

        # The mongo collection of document documents
        self.documents = self.database['documents']

    def get_document_model(self):
        """
        Return the class for the document model.
        """
        from kolibri.document import Document

        # Create a datastore-aware document
        document = Document
        document.storage = self

        return document

    def count(self):
        return self.documents.count()

    def mongo_to_object(self, document_data):
        """
        Return Document object when given texts
        returned from Mongo DB.
        """
        Document = self.get_model('document')

        document_data['id'] = document_data['_id']

        return Document(**document_data)

    def filter(self, **kwargs):
        """
        Returns a list of documents in the database
        that match the parameters specified.
        """
        import pymongo

        page_size = kwargs.pop('page_size', 1000)
        order_by = kwargs.pop('order_by', None)
        tags = kwargs.pop('tags', [])
        exclude_text = kwargs.pop('exclude_text', None)
        exclude_text_words = kwargs.pop('exclude_text_words', [])
        search_text_contains = kwargs.pop('search_text_contains', None)

        if tags:
            kwargs['tags'] = {
                '$in': tags
            }

        if exclude_text:
            if 'text' not in kwargs:
                kwargs['text'] = {}
            elif 'text' in kwargs and isinstance(kwargs['text'], str):
                text = kwargs.pop('text')
                kwargs['text'] = {
                    '$eq': text
                }
            kwargs['text']['$nin'] = exclude_text

        if exclude_text_words:
            if 'text' not in kwargs:
                kwargs['text'] = {}
            elif 'text' in kwargs and isinstance(kwargs['text'], str):
                text = kwargs.pop('text')
                kwargs['text'] = {
                    '$eq': text
                }
            exclude_word_regex = '|'.join([
                '.*{}.*'.format(word) for word in exclude_text_words
            ])
            kwargs['text']['$not'] = re.compile(exclude_word_regex)

        if search_text_contains:
            or_regex = '|'.join([
                '{}'.format(re.escape(word)) for word in search_text_contains.split(' ')
            ])
            kwargs['search_text'] = re.compile(or_regex)

        mongo_ordering = []

        if order_by:

            # Sort so that newer datetimes appear first
            if 'created_at' in order_by:
                order_by.remove('created_at')
                mongo_ordering.append(('created_at', pymongo.DESCENDING,))

            for order in order_by:
                mongo_ordering.append((order, pymongo.ASCENDING))

        total_documents = self.documents.find(kwargs).count()

        for start_index in range(0, total_documents, page_size):
            if mongo_ordering:
                for match in self.documents.find(kwargs).sort(mongo_ordering).skip(start_index).limit(page_size):
                    yield self.mongo_to_object(match)
            else:
                for match in self.documents.find(kwargs).skip(start_index).limit(page_size):
                    yield self.mongo_to_object(match)

    def create(self, **kwargs):
        """
        Creates a new document matching the keyword arguments specified.
        Returns the created document.
        """
        Document = self.get_model('document')

        if 'tags' in kwargs:
            kwargs['tags'] = list(set(kwargs['tags']))

        if 'search_text' not in kwargs:
            kwargs['search_text'] = self.tagger.get_text_index_string(kwargs['text'])

        if 'search_target_text' not in kwargs:
            if kwargs.get('in_response_to'):
                kwargs['search_target_text'] = self.tagger.get_text_index_string(kwargs['in_response_to'])

        inserted = self.documents.insert_one(kwargs)

        kwargs['id'] = inserted.inserted_id

        return Document(**kwargs)

    def create_many(self, documents):
        """
        Creates multiple document entries.
        """
        create_documents = []

        for document in documents:
            document_data = document.serialize()
            tag_data = list(set(document_data.pop('tags', [])))
            document_data['tags'] = tag_data

            if not document.search_text:
                document_data['search_text'] = self.tagger.get_text_index_string(document.text)

            if not document.search_target_text and document.in_response_to:
                document_data['search_target_text'] = self.tagger.get_text_index_string(document.in_response_to)

            create_documents.append(document_data)

        self.documents.insert_many(create_documents)

    def update(self, document):
        data = document.serialize()
        data.pop('id', None)
        data.pop('tags', None)

        data['search_text'] = self.tagger.get_text_index_string(data['text'])

        if data.get('in_response_to'):
            data['search_target_text'] = self.tagger.get_text_index_string(data['in_response_to'])

        update_data = {
            '$set': data
        }

        if document.tags:
            update_data['$addToSet'] = {
                'tags': {
                    '$each': document.tags
                }
            }

        search_parameters = {}

        if document.id is not None:
            search_parameters['_id'] = document.id
        else:
            search_parameters['text'] = document.text
            search_parameters['document'] = document.document

        update_operation = self.documents.update_one(
            search_parameters,
            update_data,
            upsert=True
        )

        if update_operation.acknowledged:
            document.id = update_operation.upserted_id

        return document

    def get_random(self):
        """
        Returns a random document from the database
        """
        from random import randint

        count = self.count()

        if count < 1:
            raise Exception('database is empty')

        random_integer = randint(0, count - 1)

        documents = self.documents.find().limit(1).skip(random_integer)

        return self.mongo_to_object(list(documents)[0])

    def remove(self, document_text):
        """
        Removes the document that matches the input text.
        """
        self.documents.delete_one({'text': document_text})

    def drop(self):
        """
        Remove the database.
        """
        self.client.drop_database(self.database.name)
